# -*- coding: utf-8 -*-
"""The OpenAI token counting class. The token calculation of vision models
follows
https://platform.openai.com/docs/guides/images-vision?api-mode=chat#calculating-costs
"""
import base64
import io
import json
import math
from http import HTTPStatus
from typing import Any

import requests

from ._token_base import TokenCounterBase


def _calculate_tokens_for_high_quality_image(
    base_tokens: int,
    tile_tokens: int,
    width: int,
    height: int,
) -> int:
    """Calculate the number of tokens for a high-quality image, which follows
    https://platform.openai.com/docs/guides/images-vision?api-mode=chat#calculating-costs
    """
    # Step1: scale to fit within a 2048x2048 box
    if width > 2048 or height > 2048:
        ratio = min(2048 / width, 2048 / height)
        width = int(width * ratio)
        height = int(height * ratio)

    # Step2: Scale to make the shortest side 768 pixels
    shortest_side = min(width, height)
    if shortest_side != 768:
        ratio = 768 / shortest_side
        width = int(width * ratio)
        height = int(height * ratio)

    # Step3: Calculate how many 512px tiles are needed
    tiles_width = (width + 511) // 512
    tiles_height = (height + 511) // 512
    total_tiles = tiles_width * tiles_height

    # Step4: Calculate the total tokens
    total_tokens = (total_tiles * tile_tokens) + base_tokens

    return total_tokens


def _get_size_of_image_url(url: str) -> tuple[int, int]:
    """Get the size of an image from the given URL.

    Args:
        url (`str`):
            A web URL or base64 encoded image URL.

    Returns:
        `tuple[int, int]`:
            A tuple containing the width and height of the image.
    """
    if url.startswith("data:image/"):
        base64_data = url.split("base64,")[1]
        image_data = base64.b64decode(base64_data)

    else:
        response = None
        for _ in range(3):
            response = requests.get(url)
            if response.status_code == HTTPStatus.OK:
                break
        response.raise_for_status()
        image_data = response.content

    from PIL import Image

    image = Image.open(io.BytesIO(image_data))
    width, height = image.size
    return width, height


def _get_base_and_tile_tokens(model_name: str) -> tuple[int, int]:
    """Get the base and tile tokens for the given OpenAI model.

    Args:
        model_name (`str`):
            The name of the model.

    Returns:
        `tuple[int, int]`:
            A tuple containing the base tokens and tile tokens.
    """
    if any(
        model_name.startswith(_)
        for _ in [
            "gpt-4o",
            "gpt-4.1",
            "gpt-4.5",
        ]
    ):
        return 85, 170

    if any(
        model_name.startswith(_)
        for _ in [
            "o1",
            "o1-pro",
            "o3",
        ]
    ):
        return 75, 150

    if model_name.startswith("4o-mini"):
        return 2833, 5667

    raise ValueError(
        f"Unsupported OpenAI model {model_name} for token counting. ",
    )


def _calculate_tokens_for_tools(
    model_name: str,
    tools: list[dict],
    encoding: Any,
) -> int:
    """Calculate the tokens for the given tools JSON schema, which follows the
    OpenAI cookbook
    https://github.com/openai/openai-cookbook/blob/6dfb7920b59a45291f7df4ea41338d1faf9ef1e8/examples/How_to_count_tokens_with_tiktoken.ipynb
    """
    if not tools:
        return 0

    func_init = 10
    prop_init = 3
    prop_key = 3
    enum_init = -3
    enum_item = 3
    func_end = 12

    if model_name.startswith("gpt-4o"):
        func_init = 7

    func_token_count = 0
    for f in tools:
        func_token_count += func_init
        function = f["function"]
        f_name = function["name"]
        f_desc = function.get("description", "").removesuffix(".")
        func_token_count += len(encoding.encode(f"{f_name}:{f_desc}"))

        properties = function["parameters"]["properties"]

        if len(properties) > 0:
            func_token_count += prop_init
            for key in properties.keys():
                func_token_count += prop_key
                p_name = key
                p_type = properties[key]["type"]
                p_desc = (
                    properties[key].get("description", "").removesuffix(".")
                )

                if "enum" in properties[key].keys():
                    func_token_count += enum_init
                    for item in properties[key]["enum"]:
                        func_token_count += enum_item
                        func_token_count += len(encoding.encode(item))

                func_token_count += len(
                    encoding.encode(f"{p_name}:{p_type}:{p_desc}"),
                )
    func_token_count += func_end

    return func_token_count


def _count_content_tokens_for_openai_vision_model(
    model_name: str,
    content: list[dict],
    encoding: Any,
) -> int:
    """Yield the number of tokens for the content of an OpenAI vision model.
    Implemented according to https://platform.openai.com/docs/guides/vision.

    Args:
        model_name (`str`):
            The name of the model.
        content (`list[dict]`):
            A list of dictionaries.
        encoding (`Any`):
            The encoding object.

    Example:
        .. code-block:: python

            _yield_tokens_for_openai_vision_model(
                [
                    {
                        "type": "text",
                        "text": "xxx",
                    },
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": "xxx",
                            "detail": "auto",
                        }
                    },
                    # ...
                ]
            )

    Returns:
        `Generator[int, None, None]`: Generate the number of tokens in a
        generator.
    """

    num_tokens = 0
    for item in content:
        assert isinstance(item, dict), (
            "The content field should be a list of dictionaries, but got "
            f"{type(item)}."
        )

        typ = item.get("type", None)
        if typ == "text":
            num_tokens += len(
                encoding.encode(item["text"]),
            )

        elif typ == "image_url":
            width, height = _get_size_of_image_url(item["image_url"]["url"])

            # Different counting logic for different models
            if any(
                model_name.startswith(_)
                for _ in [
                    "gpt-4.1-mini",
                    "gpt-4.1-nano",
                    "o4-mini",
                ]
            ):
                patches = min(
                    math.ceil(width / 32) * math.ceil(height / 32),
                    1536,
                )
                if model_name.startswith("gpt-4.1-mini"):
                    num_tokens += math.ceil(patches * 1.62)

                elif model_name.startswith("gpt-4.1-nano"):
                    num_tokens += math.ceil(patches * 2.46)

                else:
                    num_tokens += math.ceil(patches * 1.72)

            elif any(
                model_name.startswith(_)
                for _ in [
                    "gpt-4o",
                    "gpt-4.1",
                    "gpt-4o-mini",
                    "o",
                ]
            ):
                base_tokens, tile_tokens = _get_base_and_tile_tokens(
                    model_name,
                )

                # By default, we use high here to avoid undercounting tokens
                detail = item.get("image_url").get("detail", "high")
                if detail == "low":
                    num_tokens += base_tokens

                elif detail in ["auto", "high"]:
                    num_tokens += _calculate_tokens_for_high_quality_image(
                        base_tokens,
                        tile_tokens,
                        width,
                        height,
                    )

                else:
                    raise ValueError(
                        f"Unsupported image detail {detail}, expected "
                        f"one of ['low', 'auto', 'high'].",
                    )

        else:
            raise ValueError(
                "The type field currently only supports 'text' "
                f"and 'image_url', but got {typ}.",
            )

    return num_tokens


class OpenAITokenCounter(TokenCounterBase):
    """The OpenAI token counting class."""

    def __init__(self, model_name: str) -> None:
        """Initialize the OpenAI token counter.

        Args:
            model_name (`str`):
                The name of the OpenAI model to use for token counting.
        """
        self.model_name = model_name

    async def count(
        self,
        messages: list[dict[str, Any]],
        tools: list[dict] = None,
        **kwargs: Any,
    ) -> int:
        """Count the token numbers of the given messages.

        .. note:: OpenAI hasn't provided an official guide for counting tokens
         with tools. If you have any ideas, please open an issue on
         our GitHub repository.

        Args:
            messages (`list[dict[str, Any]]`):
                A list of dictionaries, where `role` and `content` fields are
                required.
            tools (`list[dict]`, defaults to `None`):
        """
        import tiktoken

        try:
            encoding = tiktoken.encoding_for_model(self.model_name)
        except KeyError:
            encoding = tiktoken.get_encoding("o200k_base")

        tokens_per_message = 3
        tokens_per_name = 1

        # every reply is primed with <|start|>assistant<|message|>
        num_tokens = 3
        for message in messages:
            num_tokens += tokens_per_message
            for key, value in message.items():
                # Considering vision models
                if key == "content" and isinstance(value, list):
                    num_tokens += (
                        _count_content_tokens_for_openai_vision_model(
                            self.model_name,
                            value,
                            encoding,
                        )
                    )

                elif isinstance(value, str):
                    num_tokens += len(encoding.encode(value))

                elif value is None:
                    continue

                elif key == "tool_calls":
                    # TODO: This is only a temporary solution, since OpenAI
                    # hasn't provided an official guide for counting tokens
                    # with tool results.
                    num_tokens += len(
                        encoding.encode(
                            json.dumps(value, ensure_ascii=False),
                        ),
                    )

                else:
                    raise TypeError(
                        f"Invalid type {type(value)} in the {key} field: "
                        f"{value}",
                    )

                if key == "name":
                    num_tokens += tokens_per_name

        if tools:
            num_tokens += _calculate_tokens_for_tools(
                self.model_name,
                tools,
                encoding,
            )

        return num_tokens



# -*- coding: utf-8 -*-
"""The Message API in the ACEBench evaluation."""
from datetime import datetime

from ._shared_state import SharedState


class MessageApi(SharedState):
    """The message Api in the ACEBench evaluation."""

    tool_functions: list[str] = [
        "send_message",
        "delete_message",
        "view_messages_between_users",
        "search_messages",
        "get_all_message_times_with_ids",
        "get_latest_message_id",
        "get_earliest_message_id",
    ]

    def __init__(self, share_state: dict) -> None:
        """Initialize the MessageApi with shared state."""
        super().__init__(share_state)

        # 设置六个用户
        self.max_capacity = 6
        self.user_list: dict[str, dict[str, str | int]] = {
            "Eve": {
                "user_id": "USR100",
                "phone_number": "123-456-7890",
                "occupation": "Software Engineer",
            },
            "Frank": {
                "user_id": "USR101",
                "phone_number": "234-567-8901",
                "occupation": "Data Scientist",
            },
            "Grace": {
                "user_id": "USR102",
                "phone_number": "345-678-9012",
                "occupation": "Product Manager",
            },
            "Helen": {
                "user_id": "USR103",
                "phone_number": "456-789-0123",
                "occupation": "UX Designer",
            },
            "Isaac": {
                "user_id": "USR104",
                "phone_number": "567-890-1234",
                "occupation": "DevOps Engineer",
            },
            "Jack": {
                "user_id": "USR105",
                "phone_number": "678-901-2345",
                "occupation": "Marketing Specialist",
            },
        }

        # 设置六个用户之间的短信记录
        # 信息1和reminder配合  信息2和food配合
        self.inbox: dict[int, dict[str, str | int]] = {
            1: {
                "sender_id": "USR100",
                "receiver_id": "USR101",
                "message": "Hey Frank, don't forget about our meeting on "
                "2024-06-11 at 4 PM in Conference Room 1.",
                "time": "2024-06-09",
            },
            2: {
                "sender_id": "USR101",
                "receiver_id": "USR102",
                "message": """你能帮我点一个\"玛格丽特披萨\"的外卖吗,商家是达美乐。""",
                "time": "2024-03-09",
            },
            3: {
                "sender_id": "USR102",
                "receiver_id": "USR103",
                "message": "帮我查一些喜茶有哪些奶茶外卖，买一杯便宜些的奶茶。"
                "买完以后记得回复我,回复的内容是（已经买好了）",
                "time": "2023-12-05",
            },
            4: {
                "sender_id": "USR103",
                "receiver_id": "USR102",
                "message": "No problem Helen, I can assist you.",
                "time": "2024-09-09",
            },
            5: {
                "sender_id": "USR104",
                "receiver_id": "USR105",
                "message": "Isaac, are you available for a call?",
                "time": "2024-06-06",
            },
            6: {
                "sender_id": "USR105",
                "receiver_id": "USR104",
                "message": "Yes Jack, let's do it in 30 minutes.",
                "time": "2024-01-15",
            },
        }

        self.message_id_counter: int = 6

    def get_state_dict(self) -> dict:
        """Get the current state dict of the MessageApi."""

        # To avoid the error in ACEBench dataset
        inbox_state = {}
        for key, value in self.inbox.items():
            inbox_state[str(key)] = value

        return {
            "MessageApi": {
                "inbox": inbox_state,
            },
        }

    def send_message(
        self,
        sender_name: str,
        receiver_name: str,
        message: str,
    ) -> dict[str, bool | str]:
        """将一条消息从一个用户发送给另一个用户。

        Args:
            sender_name (`str`):
                发送消息的用户姓名。
            receiver_name (`str`):
                接收消息的用户姓名。
            message (`str`):
                要发送的消息内容。
        """
        if not self.logged_in:
            return {"status": False, "message": "device未登录，无法发送短信"}

        if not self.wifi:
            return {"status": False, "message": "wifi关闭，此时不能发送信息"}

        if len(self.inbox) >= self.max_capacity:
            return {
                "status": False,
                "message": "内存容量不够了，你需要询问user删除哪一条短信。",
            }

        # 验证发送者和接收者是否存在
        if (
            sender_name not in self.user_list
            or receiver_name not in self.user_list
        ):
            return {"status": False, "message": "发送者或接收者不存在"}

        sender_id = self.user_list[sender_name]["user_id"]
        receiver_id = self.user_list[receiver_name]["user_id"]

        # 将短信添加到inbox
        self.message_id_counter += 1
        self.inbox[self.message_id_counter] = {
            "sender_id": sender_id,
            "receiver_id": receiver_id,
            "message": message,
        }

        return {"status": True, "message": f"短信成功发送给{receiver_name}。"}

    def delete_message(self, message_id: int) -> dict[str, bool | str]:
        """根据消息 ID 删除一条消息。

        Args:
            message_id (`int`):
                要删除的消息的 ID。
        """
        if not self.logged_in:
            return {"status": False, "message": "device未登录，无法删除短信"}
        if message_id not in self.inbox:
            return {"status": False, "message": "短信ID不存在"}

        del self.inbox[message_id]
        return {"status": True, "message": f"短信ID {message_id} 已成功删除。"}

    def view_messages_between_users(
        self,
        sender_name: str,
        receiver_name: str,
    ) -> dict:
        """获取特定用户发送给另一个用户的所有消息。

        Args:
            sender_name (`str`):
                发送消息的用户姓名。
            receiver_name (`str`):
                接收消息的用户姓名。
        """
        if not self.logged_in:
            return {
                "status": False,
                "message": "device未登录，无法查看短信信息",
            }

        if sender_name not in self.user_list:
            return {"status": False, "message": "发送者不存在"}

        if receiver_name not in self.user_list:
            return {"status": False, "message": "接收者不存在"}

        sender_id = self.user_list[sender_name]["user_id"]
        receiver_id = self.user_list[receiver_name]["user_id"]
        messages_between_users = []

        # 遍历 inbox，找出 sender_id 发送给 receiver_id 的短信
        for msg_id, msg_data in self.inbox.items():
            if (
                msg_data["sender_id"] == sender_id
                and msg_data["receiver_id"] == receiver_id
            ):
                messages_between_users.append(
                    {
                        "id": msg_id,
                        "sender": sender_name,
                        "receiver": receiver_name,
                        "message": msg_data["message"],
                    },
                )

        if not messages_between_users:
            return {"status": False, "message": "没有找到相关的短信记录"}

        return {"status": True, "messages": messages_between_users}

    def search_messages(
        self,
        user_name: str,
        keyword: str,
    ) -> dict:
        """搜索特定用户消息中包含特定关键字的消息。

        Args:
            user_name (`str`):
                要搜索消息的用户姓名。
            keyword (`str`):
                要在消息中搜索的关键字。
        """
        if user_name not in self.user_list:
            return {"status": False, "message": "用户不存在"}

        user_id = self.user_list[user_name]["user_id"]
        matched_messages = []

        # 遍历 inbox，找到发送或接收中包含关键词的消息
        for msg_id, msg_data in self.inbox.items():
            if (
                user_id in (msg_data["sender_id"], msg_data["receiver_id"])
                and keyword.lower() in msg_data["message"].lower()
            ):
                matched_messages.append(
                    {
                        "id": msg_id,
                        "sender_id": msg_data["sender_id"],
                        "receiver_id": msg_data["receiver_id"],
                        "message": msg_data["message"],
                    },
                )

        if not matched_messages:
            return {"status": False, "message": "没有找到包含关键词的短信"}

        return {"status": True, "messages": matched_messages}

    def get_all_message_times_with_ids(
        self,
    ) -> dict:
        """获取所有短信的时间以及对应的短信编号。"""
        if not self.logged_in:
            return {
                "status": False,
                "message": "device未登录，获取所有短信的时间以及对应的短信编号。",
            }
        message_times_with_ids = {
            msg_id: msg_data["time"] for msg_id, msg_data in self.inbox.items()
        }
        return message_times_with_ids

    def get_latest_message_id(self) -> dict:
        """获取最近发送的消息的 ID。"""
        if not self.logged_in:
            return {
                "status": False,
                "message": "device未登录，无法获取最新发送的短信ID。",
            }
        if not self.inbox:
            return {"status": False, "message": "短信记录为空"}

        # 遍历所有短信，找出时间最新的短信
        latest_message_id = None
        latest_time = None

        for message_id, message_data in self.inbox.items():
            message_time = datetime.strptime(
                str(message_data["time"]),
                "%Y-%m-%d",
            )
            if latest_time is None or message_time > latest_time:
                latest_time = message_time
                latest_message_id = message_id

        return {
            "status": True,
            "message": f"最新的短信ID是 {latest_message_id}",
            "message_id": latest_message_id,
        }

    def get_earliest_message_id(self) -> dict:
        """获取最早发送的消息的 ID。"""
        if not self.logged_in:
            return {
                "status": False,
                "message": "device未登录，无法获取最早发送的短信ID",
            }
        if not self.inbox:
            return {"status": False, "message": "短信记录为空"}

        # 遍历所有短信，找出时间最早的短信
        earliest_message_id = None
        earliest_time = None

        for message_id, message_data in self.inbox.items():
            message_time = datetime.strptime(
                str(message_data["time"]),
                "%Y-%m-%d",
            )
            if earliest_time is None or message_time < earliest_time:
                earliest_time = message_time
                earliest_message_id = message_id

        return {
            "status": True,
            "message": f"最早的短信ID是 {earliest_message_id}",
            "message_id": earliest_message_id,
        }



# -*- coding: utf-8 -*-
"""Tool memory implementation using ReMe library.

This module provides a tool memory implementation that integrates
with the ReMe library to record tool execution results and retrieve
tool usage guidelines.

Requirements:
    Python 3.12 or greater is required to use ReMe.
"""
from typing import Any

from ._reme_long_term_memory_base import ReMeLongTermMemoryBase
from ..._logging import logger
from ...message import Msg, TextBlock
from ...tool import ToolResponse


class ReMeToolLongTermMemory(ReMeLongTermMemoryBase):
    """Tool memory implementation using ReMe library.

    Tool memory records tool execution results and generates usage
    guidelines from the execution history.

    Requirements:
        Python 3.12 or greater is required to use ReMe.
    """

    async def record_to_memory(
        self,
        thinking: str,
        content: list[str],
        **kwargs: Any,
    ) -> ToolResponse:
        """Record tool execution results to build tool usage patterns.

        Record tool execution results to build a knowledge base of tool
        usage patterns.

        Use this function after successfully using tools to capture
        execution details, results, and performance metrics. Over time,
        this builds comprehensive usage guidelines and best practices
        for each tool.

        When to record:

        - After successfully executing any tool
        - After tool failures (to learn what doesn't work)
        - When discovering effective parameter combinations
        - After noteworthy tool usage patterns

        What to record: Each tool execution should include complete
        execution details.

        Args:
            thinking (`str`):
                Your reasoning about why this tool execution is worth
                recording. Mention what worked well, what could be
                improved, or lessons learned.
            content (`list[str]`):
                List of JSON strings, each representing a tool execution.
                Each JSON must have these fields:
                - create_time: Timestamp in format "YYYY-MM-DD HH:MM:SS"
                - tool_name: Name of the tool executed
                - input: Input parameters as a dict
                - output: Tool's output as a string
                - token_cost: Token cost (integer)
                - success: Whether execution succeeded (boolean)
                - time_cost: Execution time in seconds (float)

                Example: '{"create_time": "2024-01-01 10:00:00",
                "tool_name": "search", "input": {"query": "Python"},
                "output": "Found 10 results", "token_cost": 100,
                "success": true, "time_cost": 1.2}'
            **kwargs (`Any`):
                Additional keyword arguments for the recording operation.

        Returns:
            `ToolResponse`:
                Confirmation message with number of executions recorded
                and guidelines generated.
        """
        logger.info(
            "[ReMeToolMemory] Entering record_to_memory - "
            "thinking: %s, content: %s, kwargs: %s",
            thinking,
            content,
            kwargs,
        )

        if not self._app_started:
            raise RuntimeError(
                "ReMeApp context not started. "
                "Please use 'async with' to initialize the app.",
            )

        try:
            import json

            # Parse each content item as a tool_call_result
            tool_call_results = []
            tool_names_set = set()

            for item in content:
                try:
                    # Parse JSON string to dict
                    tool_call_result = json.loads(item)
                    tool_call_results.append(tool_call_result)

                    # Track tool names for summary
                    if "tool_name" in tool_call_result:
                        tool_names_set.add(tool_call_result["tool_name"])

                except json.JSONDecodeError as e:
                    # Skip invalid JSON items
                    import warnings

                    warnings.warn(
                        f"Failed to parse tool call result JSON: {item}. "
                        f"Error: {str(e)}",
                    )
                    continue

            if not tool_call_results:
                return ToolResponse(
                    content=[
                        TextBlock(
                            type="text",
                            text="No valid tool call results to record.",
                        ),
                    ],
                )

            # First, add the tool call results
            await self.app.async_execute(
                name="add_tool_call_result",
                workspace_id=self.workspace_id,
                tool_call_results=tool_call_results,
                **kwargs,
            )

            # Then, summarize the tool memory for the affected tools
            if tool_names_set:
                tool_names_list = list(tool_names_set)
                await self.app.async_execute(
                    name="summary_tool_memory",
                    workspace_id=self.workspace_id,
                    tool_names=tool_names_list,
                    **kwargs,
                )

            num_results = len(tool_call_results)
            summary_text = (
                f"Successfully recorded {num_results} tool execution "
                f"result{'s' if num_results > 1 else ''} and generated "
                f"usage guidelines."
            )

            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=summary_text,
                    ),
                ],
            )

        except Exception as e:
            logger.exception("Error recording tool memory: %s", str(e))
            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"Error recording tool memory: {str(e)}",
                    ),
                ],
            )

    async def retrieve_from_memory(
        self,
        keywords: list[str],
        **kwargs: Any,
    ) -> ToolResponse:
        """Retrieve usage guidelines and best practices for tools.

        Retrieve usage guidelines and best practices for specific tools.

        .. note:: You should call this function BEFORE using a tool,
         especially if you're uncertain about its proper usage or want to
         follow established best practices. This retrieves synthesized
         guidelines based on past tool executions.

        Use this when:

        - About to use a tool and want to know the best practices
        - Uncertain about tool parameters or usage patterns
        - Want to learn from past successful/failed tool executions
        - User asks "how should I use this tool?" or "what's the best
          way to..."
        - Need to understand tool performance characteristics or
          limitations

        Benefits of retrieving first:

        - Learn from accumulated tool usage experience
        - Avoid common mistakes and pitfalls
        - Use optimal parameter combinations
        - Understand tool performance and cost characteristics
        - Follow established best practices

        Args:
            keywords (`list[str]`):
                List of tool names to retrieve guidelines for. Use the
                exact tool names. Examples: ["search"],
                ["database_query", "cache_get"], ["api_call"].
            **kwargs (`Any`):
                Additional keyword arguments for the retrieval operation.

        Returns:
            `ToolResponse`:
                Retrieved usage guidelines and best practices for the
                specified tools. If no guidelines exist yet, you'll
                receive a message indicating that.
        """
        logger.info(
            "[ReMeToolMemory] Entering retrieve_from_memory - "
            "keywords: %s, kwargs: %s",
            keywords,
            kwargs,
        )

        if not self._app_started:
            raise RuntimeError(
                "ReMeApp context not started. "
                "Please use 'async with' to initialize the app.",
            )

        try:
            # Join all tool names with comma
            tool_names = ",".join(keywords)

            # Retrieve tool guidelines for all tools at once
            result = await self.app.async_execute(
                name="retrieve_tool_memory",
                workspace_id=self.workspace_id,
                tool_names=tool_names,
                **kwargs,
            )

            # Extract the answer from the result
            answer = result.get("answer", "")
            if answer:
                combined_text = answer
            else:
                combined_text = f"No tool guidelines found for: {tool_names}"

            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=combined_text,
                    ),
                ],
            )

        except Exception as e:
            logger.exception("Error retrieving tool memory: %s", str(e))
            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"Error retrieving tool memory: {str(e)}",
                    ),
                ],
            )

    def _extract_content_from_messages(self, msg_list: list[Msg]) -> list[str]:
        """Extract content strings from messages.

        Args:
            msg_list (`list[Msg]`):
                List of messages to extract content from.

        Returns:
            `list[str]`:
                List of extracted content strings.
        """
        content_list = []
        for msg in msg_list:
            if isinstance(msg.content, str):
                content_list.append(msg.content)
            elif isinstance(msg.content, list):
                content_list.extend(
                    self._extract_text_from_blocks(msg.content),
                )
        return content_list

    def _extract_text_from_blocks(self, blocks: list) -> list[str]:
        """Extract text from content blocks.

        Args:
            blocks (`list`):
                List of content blocks.

        Returns:
            `list[str]`:
                List of extracted text strings.
        """
        texts = []
        for block in blocks:
            if isinstance(block, dict) and block.get("type") == "text":
                texts.append(block.get("text", ""))
            elif isinstance(block, str):
                texts.append(block)
        return texts

    def _parse_tool_call_results(
        self,
        content_list: list[str],
    ) -> tuple[list[dict], set[str]]:
        """Parse JSON content strings into tool call results.

        Args:
            content_list (`list[str]`):
                List of JSON strings to parse.

        Returns:
            `tuple[list[dict], set[str]]`:
                Tuple of (tool_call_results, tool_names_set).
        """
        import json
        import warnings

        tool_call_results = []
        tool_names_set = set()

        for item in content_list:
            try:
                tool_call_result = json.loads(item)
                tool_call_results.append(tool_call_result)
                if "tool_name" in tool_call_result:
                    tool_names_set.add(tool_call_result["tool_name"])
            except json.JSONDecodeError as e:
                warnings.warn(
                    f"Failed to parse tool call result JSON: {item}. "
                    f"Error: {str(e)}",
                )

        return tool_call_results, tool_names_set

    async def record(
        self,
        msgs: list[Msg | None],
        **kwargs: Any,
    ) -> None:
        """Record the content to the tool memory.

        This method extracts content from messages and treats them as
        JSON strings representing tool_call_results, similar to
        record_to_memory.

        Args:
            msgs (`list[Msg | None]`):
                The messages to record to memory. Each message's content
                should be a JSON string or list of JSON strings
                representing tool_call_results.
            **kwargs (`Any`):
                Additional keyword arguments for the recording.
        """
        if isinstance(msgs, Msg):
            msgs = [msgs]

        # Filter out None
        msg_list = [_ for _ in msgs if _]
        if not msg_list:
            return

        if not all(isinstance(_, Msg) for _ in msg_list):
            raise TypeError(
                "The input messages must be a list of Msg objects.",
            )

        if not self._app_started:
            raise RuntimeError(
                "ReMeApp context not started. "
                "Please use 'async with' to initialize the app.",
            )

        try:
            # Extract content from messages and parse as tool_call_results
            content_list = self._extract_content_from_messages(msg_list)
            if not content_list:
                return

            # Parse each content item as a tool_call_result
            tool_call_results, tool_names_set = self._parse_tool_call_results(
                content_list,
            )
            if not tool_call_results:
                return

            # First, add the tool call results
            await self.app.async_execute(
                name="add_tool_call_result",
                workspace_id=self.workspace_id,
                tool_call_results=tool_call_results,
                **kwargs,
            )

            # Then, summarize the tool memory for the affected tools
            if tool_names_set:
                tool_names_list = list(tool_names_set)
                await self.app.async_execute(
                    name="summary_tool_memory",
                    workspace_id=self.workspace_id,
                    tool_names=tool_names_list,
                    **kwargs,
                )

        except Exception as e:
            # Log the error but don't raise to maintain compatibility
            logger.exception(
                "Error recording tool messages to memory: %s",
                str(e),
            )
            import warnings

            warnings.warn(
                f"Error recording tool messages to memory: {str(e)}",
            )

    def _extract_tool_names_from_message(self, msg: Msg) -> str:
        """Extract tool names from a message.

        Args:
            msg (`Msg`):
                Message to extract tool names from.

        Returns:
            `str`:
                Extracted tool names as a string.
        """
        if isinstance(msg.content, str):
            return msg.content

        if isinstance(msg.content, list):
            content_parts = []
            for block in msg.content:
                if isinstance(block, dict) and "text" in block:
                    content_parts.append(block["text"])
            return " ".join(content_parts)

        return ""

    def _format_retrieve_result(self, result: Any) -> str:
        """Format the retrieve result into a string.

        Args:
            result (`Any`):
                Result from the retrieve operation.

        Returns:
            `str`:
                Formatted result string.
        """
        if isinstance(result, dict) and "answer" in result:
            return result["answer"]
        if isinstance(result, str):
            return result
        return str(result)

    async def retrieve(
        self,
        msg: Msg | list[Msg] | None,
        **kwargs: Any,
    ) -> str:
        """Retrieve tool guidelines from memory.

        Retrieve tool guidelines from memory based on message content.

        Args:
            msg (`Msg | list[Msg] | None`):
                The message containing tool names or queries to
                retrieve guidelines for.
            **kwargs (`Any`):
                Additional keyword arguments.

        Returns:
            `str`:
                The retrieved tool guidelines as a string.
        """
        if msg is None:
            return ""

        if isinstance(msg, Msg):
            msg = [msg]

        if not isinstance(msg, list) or not all(
            isinstance(_, Msg) for _ in msg
        ):
            raise TypeError(
                "The input message must be a Msg or a list of Msg objects.",
            )

        if not self._app_started:
            raise RuntimeError(
                "ReMeApp context not started. "
                "Please use 'async with' to initialize the app.",
            )

        try:
            # Extract tool names from the last message
            last_msg = msg[-1]
            tool_names = self._extract_tool_names_from_message(last_msg)

            if not tool_names:
                return ""

            # Retrieve tool guidelines
            result = await self.app.async_execute(
                name="retrieve_tool_memory",
                workspace_id=self.workspace_id,
                tool_names=tool_names,
                **kwargs,
            )

            return self._format_retrieve_result(result)

        except Exception as e:
            logger.exception("Error retrieving tool guidelines: %s", str(e))
            import warnings

            warnings.warn(f"Error retrieving tool guidelines: {str(e)}")
            return ""



# -*- coding: utf-8 -*-
"""The formatter module."""

from abc import abstractmethod
from typing import Any, List

from .._utils._common import _save_base64_data
from ..message import Msg, AudioBlock, ImageBlock, TextBlock


class FormatterBase:
    """The base class for formatters."""

    @abstractmethod
    async def format(self, *args: Any, **kwargs: Any) -> list[dict[str, Any]]:
        """Format the Msg objects to a list of dictionaries that satisfy the
        API requirements."""

    @staticmethod
    def assert_list_of_msgs(msgs: list[Msg]) -> None:
        """Assert that the input is a list of Msg objects.

        Args:
            msgs (`list[Msg]`):
                A list of Msg objects to be validated.
        """
        if not isinstance(msgs, list):
            raise TypeError("Input must be a list of Msg objects.")

        for msg in msgs:
            if not isinstance(msg, Msg):
                raise TypeError(
                    f"Expected Msg object, got {type(msg)} instead.",
                )

    @staticmethod
    def convert_tool_result_to_string(
        output: str | List[TextBlock | ImageBlock | AudioBlock],
    ) -> str:
        """Turn the tool result list into a textual output to be compatible
        with the LLM API that doesn't support multimodal data.

        Args:
            output (`str | List[TextBlock | ImageBlock | AudioBlock]`):
                The output of the tool response, including text and multimodal
                data like images and audio.

        Returns:
            `str`:
                A string representation of the tool result, with text blocks
                concatenated and multimodal data represented by file paths
                or URLs.
        """

        if isinstance(output, str):
            return output

        textual_output = []
        for block in output:
            assert isinstance(block, dict) and "type" in block, (
                f"Invalid block: {block}, a TextBlock, ImageBlock, or "
                f"AudioBlock is expected."
            )
            if block["type"] == "text":
                textual_output.append(block["text"])

            elif block["type"] in ["image", "audio", "video"]:
                assert "source" in block, (
                    f"Invalid {block['type']} block: {block}, 'source' key "
                    "is required."
                )
                source = block["source"]
                # Save the image locally and return the file path
                if source["type"] == "url":
                    textual_output.append(
                        f"The returned {block['type']} can be found "
                        f"at: {source['url']}",
                    )

                elif source["type"] == "base64":
                    path_temp_file = _save_base64_data(
                        source["media_type"],
                        source["data"],
                    )
                    textual_output.append(
                        f"The returned {block['type']} can be found "
                        f"at: {path_temp_file}",
                    )

                else:
                    raise ValueError(
                        f"Invalid image source: {block['source']}, "
                        "expected 'url' or 'base64'.",
                    )

            else:
                raise ValueError(
                    f"Unsupported block type: {block['type']}, "
                    "expected 'text', 'image', 'audio', or 'video'.",
                )

        if len(textual_output) == 1:
            return textual_output[0]

        else:
            return "\n".join("- " + _ for _ in textual_output)



# -*- coding: utf-8 -*-
"""The types in agentscope"""

from ._hook import (
    AgentHookTypes,
    ReActAgentHookTypes,
)
from ._object import Embedding
from ._json import (
    JSONPrimitive,
    JSONSerializableObject,
)
from ._tool import ToolFunction

__all__ = [
    "AgentHookTypes",
    "ReActAgentHookTypes",
    "Embedding",
    "JSONPrimitive",
    "JSONSerializableObject",
    "ToolFunction",
]



# -*- coding: utf-8 -*-
"""Base long-term memory implementation using ReMe library.

This module provides a base class for long-term memory implementations
that integrate with the ReMe library. ReMe enables agents to maintain
persistent, searchable memories across sessions and contexts.

The module handles the integration between AgentScope's memory system and
the ReMe library, including:
- Model configuration and API credential management
- Context lifecycle management (async context managers)
- Graceful handling of missing dependencies
- Error handling with helpful installation instructions

Key Features:
- Supports both DashScope and OpenAI model providers
- Automatic extraction of API credentials and endpoints
- Flexible configuration via config files or kwargs
- Safe fallback behavior when reme_ai is not installed

Dependencies:
    The ReMe library is an optional dependency that must be installed:

        .. code-block:: bash

            pip install reme-ai

    Python 3.12 or greater is required to use ReMe.
    For more information, visit: https://github.com/modelscope/reMe

Subclasses:
    This base class is extended by specific memory type implementations:
    - ReMeToolLongTermMemory: For tool execution patterns and guidelines
    - ReMeTaskLongTermMemory: For task execution experiences and learnings
    - ReMePersonalLongTermMemory: For user preferences and personal information

Example:
    .. code-block:: python

        from agentscope.models import OpenAIChatModel
        from agentscope.embedding import OpenAITextEmbedding
        from agentscope.memory._reme import ReMeToolLongTermMemory

        # Initialize models
        model = OpenAIChatModel(model_name="gpt-4", api_key="...")
        embedding = OpenAITextEmbedding(
            model_name="text-embedding-3-small", api_key="...")

        # Create memory instance
        memory = ReMeToolLongTermMemory(
            agent_name="my_agent",
            user_name="user_123",
            model=model,
            embedding_model=embedding
        )

        # Use memory in async context
        async with memory:
            # Record tool execution
            await memory.record_to_memory(
                thinking="This tool worked well for data processing",
                content=['{"tool_name": "process_data", "success": true, ...}']
            )

            # Retrieve tool guidelines
            result = await memory.retrieve_from_memory(
                keywords=["process_data"]
            )

"""
from abc import ABCMeta
from typing import Any

from .._long_term_memory_base import LongTermMemoryBase
from ...embedding import (
    DashScopeTextEmbedding,
    OpenAITextEmbedding,
)
from ...model import (
    DashScopeChatModel,
    OpenAIChatModel,
)


class ReMeLongTermMemoryBase(LongTermMemoryBase, metaclass=ABCMeta):
    """Base class for ReMe-based long-term memory implementations.

    This class provides the foundation for integrating AgentScope with the ReMe
    library, enabling agents to maintain and retrieve long-term memories across
    different contexts.

    The ReMe library must be installed separately:
        pip install reme-ai

    Requirements:
        Python 3.12 or greater is required to use ReMe.

    If the library is not installed, a warning will be issued during
    initialization,
    and runtime errors with installation instructions will be raised
    when attempting
    to use memory operations.
    """

    def __init__(
        self,
        agent_name: str | None = None,
        user_name: str | None = None,
        run_name: str | None = None,
        model: DashScopeChatModel | OpenAIChatModel | None = None,
        embedding_model: (
            DashScopeTextEmbedding | OpenAITextEmbedding | None
        ) = None,
        reme_config_path: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the ReMe-based long-term memory.

        This constructor sets up the connection to the ReMe
        library and configures
        the necessary models for memory operations. The ReMe app
        will be initialized
        with the provided model configurations.

        Args:
            agent_name (`str | None`, optional):
                Name identifier for the agent. Used for organizing
                memories by agent.
            user_name (`str | None`, optional):
                Unique identifier for the user or workspace. This maps
                to workspace_id in ReMe and helps isolate memories across
                different users/workspaces.
            run_name (`str | None`, optional):
                Name identifier for the current execution run or session.
            model (`DashScopeChatModel | OpenAIChatModel | None`, optional):
                The chat model to use for memory operations. The model's
                API credentials and endpoint will be extracted and
                passed to ReMe.
            embedding_model (`DashScopeTextEmbedding | OpenAITextEmbedding | \
            None`, optional):
                The embedding model to use for semantic memory retrieval.
                The model's API credentials and endpoint will be
                extracted and passed to ReMe.
            reme_config_path (`str | None`, optional):
                Path to a custom ReMe configuration file. If not provided, ReMe
                will use its default configuration.
            **kwargs (`Any`):
                Additional keyword arguments to pass to the
                ReMeApp constructor.
                These can include custom ReMe configuration parameters.

        Raises:
            `ValueError`:
                If the provided model is not a DashScopeChatModel or
                OpenAIChatModel, or if the embedding_model is not a
                DashScopeTextEmbedding or OpenAITextEmbedding.

        Note:
            If the reme_ai library is not installed, a warning will be
            issued and self.app will be set to None. Subsequent memory
            operations will raise RuntimeError with installation
            instructions.

        Example:
            .. code-block:: python

                from agentscope.models import OpenAIChatModel
                from agentscope.embedding import OpenAITextEmbedding
                from agentscope.memory._reme import ReMeToolLongTermMemory

                # Initialize models
                model = OpenAIChatModel(
                    model_name="gpt-4",
                    api_key="your-api-key"
                )
                embedding = OpenAITextEmbedding(
                    model_name="text-embedding-3-small",
                    api_key="your-api-key"
                )

                # Create memory instance
                memory = ReMeToolLongTermMemory(
                    agent_name="my_agent",
                    user_name="user_123",
                    run_name="session_001",
                    model=model,
                    embedding_model=embedding
                )

                # Use with async context manager
                async with memory:
                    # Memory operations...
                    pass

        """
        super().__init__()

        # Store agent and workspace identifiers
        self.agent_name = agent_name
        # Maps to ReMe's workspace_id concept
        self.workspace_id = user_name
        self.run_name = run_name

        # Build configuration arguments for ReMeApp
        # These will be passed as command-line style config overrides
        config_args = []

        # Extract LLM API credentials based on model type
        # DashScope uses a fixed endpoint, OpenAI can have custom base_url
        if isinstance(model, DashScopeChatModel):
            llm_api_base = "https://dashscope.aliyuncs.com/compatible-mode/v1"
            llm_api_key = model.api_key

        elif isinstance(model, OpenAIChatModel):
            llm_api_base = str(getattr(model.client, "base_url", None))
            llm_api_key = str(getattr(model.client, "api_key", None))

        else:
            raise ValueError(
                f"model must be a DashScopeChatModel or "
                f"OpenAIChatModel instance. "
                f"Got {type(model).__name__} instead.",
            )

        # Extract model name and add to config if provided
        llm_model_name = model.model_name

        if llm_model_name:
            config_args.append(f"llm.default.model_name={llm_model_name}")

        # Extract embedding model API credentials based on type
        # Similar to LLM, DashScope uses fixed endpoint,
        # OpenAI can be customized
        if isinstance(embedding_model, DashScopeTextEmbedding):
            embedding_api_base = (
                "https://dashscope.aliyuncs.com/compatible-mode/v1"
            )
            embedding_api_key = embedding_model.api_key

        elif isinstance(embedding_model, OpenAITextEmbedding):
            embedding_api_base = getattr(
                embedding_model.client,
                "base_url",
                None,
            )
            embedding_api_key = getattr(
                embedding_model.client,
                "api_key",
                None,
            )

        else:
            raise ValueError(
                "embedding_model must be a DashScopeTextEmbedding or "
                "OpenAITextEmbedding instance. "
                f"Got {type(embedding_model).__name__} instead.",
            )

        # Extract embedding model name and add to config if provided
        embedding_model_name = embedding_model.model_name

        if embedding_model_name:
            config_args.append(
                f"embedding_model.default.model_name={embedding_model_name}",
            )

        # Attempt to import and initialize ReMe
        # If import fails, set app to None and issue a warning
        # This allows the class to be instantiated even without
        # reme_ai installed
        try:
            from reme_ai import ReMeApp
        except ImportError as e:
            raise ImportError(
                "The 'reme_ai' library is required for ReMe-based "
                "long-term memory. Please install it by `pip install reme-ai`,"
                "and visit: https://github.com/modelscope/reMe for more "
                "information.",
            ) from e

        # Initialize ReMe with extracted configurations
        self.app = ReMeApp(
            *config_args,  # Config overrides as positional args
            llm_api_key=llm_api_key,
            llm_api_base=llm_api_base,
            embedding_api_key=embedding_api_key,
            embedding_api_base=embedding_api_base,
            # Optional custom config file
            config_path=reme_config_path,
            # Additional ReMe-specific configurations
            **kwargs,
        )

        # Track if the app context is active (started via __aenter__)
        self._app_started = False

    async def __aenter__(self) -> "ReMeLongTermMemoryBase":
        """Async context manager entry point.

        This method is called when entering an async context
        (using 'async with'). It initializes the ReMe app context if
        available, enabling memory operations within the context block.

        Returns:
            `ReMeLongTermMemoryBase`:
                The memory instance itself, allowing it to be used in
                the context.

        Example:
            .. code-block:: python

                memory = ReMeToolLongTermMemory(
                    agent_name="my_agent",
                    model=model,
                    embedding_model=embedding
                )

                async with memory:
                    # Memory operations can be performed here
                    await memory.record_to_memory(
                        thinking="Recording tool usage",
                        content=[...]
                    )

        """
        if self.app is not None:
            await self.app.__aenter__()
            self._app_started = True
        return self

    async def __aexit__(
        self,
        exc_type: Any,
        exc_val: Any,
        exc_tb: Any,
    ) -> None:
        """Async context manager exit point.

        This method is called when exiting an async context (at the end
        of 'async with' block or when an exception occurs). It properly
        cleans up the ReMe app context and resources.

        Args:
            exc_type (`Any`):
                The type of exception that occurred, if any. None if no
                exception.
            exc_val (`Any`):
                The exception instance that occurred, if any. None if no
                exception.
            exc_tb (`Any`):
                The traceback object for the exception, if any. None if
                no exception.

        .. note:: This method will gracefully handle the case where self.app
         is None (reme_ai not installed) by skipping the cleanup but still
         marking the app as stopped. It will also always set _app_started
         to False, ensuring the memory state is properly reset.

        Example:
            .. code-block:: python

                async with memory:
                    try:
                        # Memory operations
                        await memory.record_to_memory(...)
                    except Exception as e:
                        # __aexit__ will be called even if an exception occurs
                        print(f"Error: {e}")
                # __aexit__ has been called and resources are cleaned up

        """
        if self.app is not None:
            await self.app.__aexit__(exc_type, exc_val, exc_tb)
        self._app_started = False



# -*- coding: utf-8 -*-
"""The base class for _metric in evaluation."""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Any

from .._utils._common import _get_timestamp
from .._utils._mixin import DictMixin
from ..types import JSONSerializableObject


@dataclass
class MetricResult(DictMixin):
    """The result of a _metric."""

    name: str
    """The metric name."""

    result: str | float | int
    """The metric result."""

    created_at: str = field(default_factory=_get_timestamp)
    """The timestamp when the metric result was created."""

    message: str | None = field(default_factory=lambda: None)
    """An optional message for the metric result, can be used to provide
    additional information or context about the result."""

    metadata: dict[str, JSONSerializableObject] | None = field(default=None)
    """Optional metadata for the metric result, can be used to store
    additional information related to the metric result."""


class MetricType(str, Enum):
    """The metric type enum."""

    CATEGORY = "category"
    """The metric result is a category, e.g. "pass" or "fail"."""

    NUMERICAL = "numerical"
    """The metric result is a numerical value, e.g. 0.95 or 100."""


class MetricBase(ABC):
    """The base class for _metric in evaluation."""

    def __init__(
        self,
        name: str,
        metric_type: MetricType,
        description: str | None = None,
        categories: list[str] | None = None,
    ) -> None:
        """Initialize the _metric object.

        Args:
            name (`str`):
                The name of the metric.
            metric_type (`MetricType`):
                The type of the metric, can be either "category" or
                "numerical", which will determine how to display the result.
            description (`str`):
                The description of the metric.
            categories (`list[str] | None`, optional):
                The candidate categories. If `metric_type` is "category", the
                categories must be provided, otherwise it should be `None`.
        """
        self.name = name
        self.metric_type = metric_type
        self.description = description

        if metric_type == MetricType.CATEGORY and categories is None:
            raise ValueError(
                "Categories must be provided for category metrics.",
            )

        self.categories = categories

    @abstractmethod
    async def __call__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> MetricResult:
        """The call function to calculate the _metric result"""



# -*- coding: utf-8 -*-
"""The evaluator module in AgentScope."""

from ._evaluator_base import EvaluatorBase
from ._ray_evaluator import RayEvaluator
from ._general_evaluator import GeneralEvaluator

__all__ = [
    "EvaluatorBase",
    "RayEvaluator",
    "GeneralEvaluator",
]



# -*- coding: utf-8 -*-
"""The memory base class."""

from abc import abstractmethod
from typing import Any

from ..message import Msg
from ..module import StateModule


class MemoryBase(StateModule):
    """The base class for memory in agentscope."""

    @abstractmethod
    async def add(self, *args: Any, **kwargs: Any) -> None:
        """Add items to the memory."""

    @abstractmethod
    async def delete(self, *args: Any, **kwargs: Any) -> None:
        """Delete items from the memory."""

    @abstractmethod
    async def retrieve(self, *args: Any, **kwargs: Any) -> None:
        """Retrieve items from the memory."""

    @abstractmethod
    async def size(self) -> int:
        """Get the size of the memory."""

    @abstractmethod
    async def clear(self) -> None:
        """Clear the memory content."""

    @abstractmethod
    async def get_memory(self, *args: Any, **kwargs: Any) -> list[Msg]:
        """Get the memory content."""

    @abstractmethod
    def state_dict(self) -> dict:
        """Get the state dictionary of the memory."""

    @abstractmethod
    def load_state_dict(self, state_dict: dict, strict: bool = True) -> None:
        """Load the state dictionary of the memory."""



# -*- coding: utf-8 -*-
"""The utility functions for text file tools in agentscope."""
from ...exception import ToolInvalidArgumentsError


def _calculate_view_ranges(
    old_n_lines: int,
    new_n_lines: int,
    start: int,
    end: int,
    extra_view_n_lines: int = 5,
) -> tuple[int, int]:
    """Calculate after writing the new content, the view ranges of the file.

    Args:
        old_n_lines (`int`):
            The number of lines before writing the new content.
        new_n_lines (`int`):
            The number of lines after writing the new content.
        start (`int`):
            The start line of the writing range.
        end (`int`):
            The end line of the writing range.
        extra_view_n_lines (`int`, optional):
            The number of extra lines to view before and after the range.
    """

    view_start = max(1, start - extra_view_n_lines)

    delta_lines = new_n_lines - old_n_lines
    view_end = min(end + delta_lines + extra_view_n_lines, new_n_lines)

    return view_start, view_end


def _assert_ranges(
    ranges: list[int],
) -> None:
    """Check if the ranges are valid.

    Raises:
        ToolInvalidArgumentsError: If the ranges are invalid.
    """
    if (
        isinstance(ranges, list)
        and len(ranges) == 2
        and all(isinstance(i, int) for i in ranges)
    ):
        start, end = ranges
        if start > end:
            raise ToolInvalidArgumentsError(
                f"InvalidArgumentError: The start line is greater than the "
                f"end line in the given range {ranges}.",
            )
    else:
        raise ToolInvalidArgumentsError(
            f"InvalidArgumentError: Invalid range format. Expected a list of "
            f"two integers, but got {ranges}.",
        )


def _view_text_file(
    file_path: str,
    ranges: list[int] | None = None,
) -> str:
    """Return the file content in the specified range with line numbers."""
    with open(file_path, "r", encoding="utf-8") as file:
        lines = file.readlines()

    if ranges:
        _assert_ranges(ranges)
        start, end = ranges

        if start > len(lines):
            raise ToolInvalidArgumentsError(
                f"InvalidArgumentError: The range '{ranges}' is out of bounds "
                f"for the file '{file_path}', which has only {len(lines)} "
                f"lines.",
            )

        view_content = [
            f"{index + start}: {line}"
            for index, line in enumerate(lines[start - 1 : end])
        ]

        return "".join(view_content)

    return "".join(f"{index + 1}: {line}" for index, line in enumerate(lines))



# -*- coding: utf-8 -*-
# pylint: disable=too-many-branches
"""The DeepSeek formatter module."""
import json
from typing import Any

from ._truncated_formatter_base import TruncatedFormatterBase
from .._logging import logger
from ..message import Msg, TextBlock, ToolUseBlock, ToolResultBlock
from ..token import TokenCounterBase


class DeepSeekChatFormatter(TruncatedFormatterBase):
    """Formatter for DeepSeek messages."""

    support_tools_api: bool = True
    """Whether support tools API"""

    support_multiagent: bool = False
    """Whether support multi-agent conversations"""

    support_vision: bool = False
    """Whether support vision data"""

    supported_blocks: list[type] = [
        TextBlock,
        # Tool use
        ToolUseBlock,
        ToolResultBlock,
    ]
    """The list of supported message blocks"""

    async def _format(
        self,
        msgs: list[Msg],
    ) -> list[dict[str, Any]]:
        """Format message objects into DeepSeek API format.

        Args:
            msgs (`list[Msg]`):
                The list of message objects to format.

        Returns:
            `list[dict[str, Any]]`:
                The formatted messages as a list of dictionaries.
        """
        self.assert_list_of_msgs(msgs)

        messages: list[dict] = []
        for msg in msgs:
            content_blocks: list = []
            tool_calls = []

            for block in msg.get_content_blocks():
                typ = block.get("type")
                if typ == "text":
                    content_blocks.append({**block})

                elif typ == "tool_use":
                    tool_calls.append(
                        {
                            "id": block.get("id"),
                            "type": "function",
                            "function": {
                                "name": block.get("name"),
                                "arguments": json.dumps(
                                    block.get("input", {}),
                                    ensure_ascii=False,
                                ),
                            },
                        },
                    )

                elif typ == "tool_result":
                    messages.append(
                        {
                            "role": "tool",
                            "tool_call_id": block.get("id"),
                            "content": self.convert_tool_result_to_string(
                                block.get("output"),  # type: ignore[arg-type]
                            ),
                            "name": block.get("name"),
                        },
                    )

                else:
                    logger.warning(
                        "Unsupported block type %s in the message, skipped.",
                        typ,
                    )
            content_msg = "\n".join(
                content.get("text", "") for content in content_blocks
            )
            msg_deepseek = {
                "role": msg.role,
                "content": content_msg or None,
            }

            if tool_calls:
                msg_deepseek["tool_calls"] = tool_calls

            if msg_deepseek["content"] or msg_deepseek.get("tool_calls"):
                messages.append(msg_deepseek)

        return messages


class DeepSeekMultiAgentFormatter(TruncatedFormatterBase):
    """
    DeepSeek formatter for multi-agent conversations, where more than
    a user and an agent are involved.
    """

    support_tools_api: bool = True
    """Whether support tools API"""

    support_multiagent: bool = True
    """Whether support multi-agent conversations"""

    support_vision: bool = False
    """Whether support vision data"""

    supported_blocks: list[type] = [
        TextBlock,
        # Tool use
        ToolUseBlock,
        ToolResultBlock,
    ]
    """The list of supported message blocks"""

    def __init__(
        self,
        conversation_history_prompt: str = (
            "# Conversation History\n"
            "The content between <history></history> tags contains "
            "your conversation history\n"
        ),
        token_counter: TokenCounterBase | None = None,
        max_tokens: int | None = None,
    ) -> None:
        """Initialize the DeepSeek multi-agent formatter.

        Args:
            conversation_history_prompt (`str`):
                The prompt to use for the conversation history section.
            token_counter (`TokenCounterBase | None`, optional):
                A token counter instance used to count tokens in the messages.
                If not provided, the formatter will format the messages
                without considering token limits.
            max_tokens (`int | None`, optional):
                The maximum number of tokens allowed in the formatted
                messages. If not provided, the formatter will not truncate
                the messages.
        """
        super().__init__(token_counter=token_counter, max_tokens=max_tokens)
        self.conversation_history_prompt = conversation_history_prompt

    async def _format_tool_sequence(
        self,
        msgs: list[Msg],
    ) -> list[dict[str, Any]]:
        """Given a sequence of tool call/result messages, format them into
        the required format for the DeepSeek API.

        Args:
            msgs (`list[Msg]`):
                The list of messages containing tool calls/results to format.

        Returns:
            `list[dict[str, Any]]`:
                A list of dictionaries formatted for the DeepSeek API.
        """
        return await DeepSeekChatFormatter().format(msgs)

    async def _format_agent_message(
        self,
        msgs: list[Msg],
        is_first: bool = True,
    ) -> list[dict[str, Any]]:
        """Given a sequence of messages without tool calls/results, format
        them into the required format for the DeepSeek API.

        Args:
            msgs (`list[Msg]`):
                A list of Msg objects to be formatted.
            is_first (`bool`, defaults to `True`):
                Whether this is the first agent message in the conversation.
                If `True`, the conversation history prompt will be included.

        Returns:
            `list[dict[str, Any]]`:
                A list of dictionaries formatted for the DeepSeek API.
        """

        if is_first:
            conversation_history_prompt = self.conversation_history_prompt
        else:
            conversation_history_prompt = ""

        # Format into required DeepSeek format
        formatted_msgs: list[dict] = []

        conversation_blocks: list = []
        accumulated_text = []
        for msg in msgs:
            for block in msg.get_content_blocks():
                if block["type"] == "text":
                    accumulated_text.append(f"{msg.name}: {block['text']}")

        if accumulated_text:
            conversation_blocks.append(
                {"text": "\n".join(accumulated_text)},
            )

        if conversation_blocks:
            if conversation_blocks[0].get("text"):
                conversation_blocks[0]["text"] = (
                    conversation_history_prompt
                    + "<history>\n"
                    + conversation_blocks[0]["text"]
                )

            else:
                conversation_blocks.insert(
                    0,
                    {
                        "text": conversation_history_prompt + "<history>\n",
                    },
                )

            if conversation_blocks[-1].get("text"):
                conversation_blocks[-1]["text"] += "\n</history>"

            else:
                conversation_blocks.append({"text": "</history>"})

        conversation_blocks_text = "\n".join(
            conversation_block.get("text", "")
            for conversation_block in conversation_blocks
        )

        user_message = {
            "role": "user",
            "content": conversation_blocks_text,
        }

        if conversation_blocks:
            formatted_msgs.append(user_message)

        return formatted_msgs



# -*- coding: utf-8 -*-
# pylint: disable=too-many-branches
"""The Anthropic formatter module."""

from typing import Any

from ._truncated_formatter_base import TruncatedFormatterBase
from .._logging import logger
from ..message import Msg, TextBlock, ImageBlock, ToolUseBlock, ToolResultBlock
from ..token import TokenCounterBase


class AnthropicChatFormatter(TruncatedFormatterBase):
    """Formatter for Anthropic messages."""

    support_tools_api: bool = True
    """Whether support tools API"""

    support_multiagent: bool = False
    """Whether support multi-agent conversations"""

    support_vision: bool = True
    """Whether support vision data"""

    supported_blocks: list[type] = [
        TextBlock,
        # Multimodal
        ImageBlock,
        # Tool use
        ToolUseBlock,
        ToolResultBlock,
    ]
    """The list of supported message blocks"""

    async def _format(
        self,
        msgs: list[Msg],
    ) -> list[dict[str, Any]]:
        """Format message objects into Anthropic API format.

        Args:
            msgs (`list[Msg]`):
                The list of message objects to format.

        Returns:
            `list[dict[str, Any]]`:
                The formatted messages as a list of dictionaries.

        .. note:: Anthropic suggests always passing all previous thinking
         blocks back to the API in subsequent calls to maintain reasoning
         continuity. For more details, please refer to
         `Anthropic's documentation
         <https://docs.anthropic.com/en/docs/build-with-claude/extended-thinking#preserving-thinking-blocks>`_.
        """
        self.assert_list_of_msgs(msgs)

        messages: list[dict] = []
        for index, msg in enumerate(msgs):
            content_blocks = []

            for block in msg.get_content_blocks():
                typ = block.get("type")
                if typ in ["thinking", "text", "image"]:
                    content_blocks.append({**block})

                elif typ == "tool_use":
                    content_blocks.append(
                        {
                            "id": block.get("id"),
                            "type": "tool_use",
                            "name": block.get("name"),
                            "input": block.get("input", {}),
                        },
                    )

                elif typ == "tool_result":
                    output = block.get("output")
                    if output is None:
                        content_value = [{"type": "text", "text": None}]
                    elif isinstance(output, list):
                        content_value = output
                    else:
                        content_value = [{"type": "text", "text": str(output)}]
                    messages.append(
                        {
                            "role": "user",
                            "content": [
                                {
                                    "type": "tool_result",
                                    "tool_use_id": block.get("id"),
                                    "content": content_value,
                                },
                            ],
                        },
                    )
                else:
                    logger.warning(
                        "Unsupported block type %s in the message, skipped.",
                        typ,
                    )

            # Claude only allow the first message to be system message
            if msg.role == "system" and index != 0:
                role = "user"
            else:
                role = msg.role

            msg_anthropic = {
                "role": role,
                "content": content_blocks or None,
            }

            # When both content and tool_calls are None, skipped
            if msg_anthropic["content"] or msg_anthropic.get("tool_calls"):
                messages.append(msg_anthropic)

        return messages


class AnthropicMultiAgentFormatter(TruncatedFormatterBase):
    """
    Anthropic formatter for multi-agent conversations, where more than
    a user and an agent are involved.
    """

    support_tools_api: bool = True
    """Whether support tools API"""

    support_multiagent: bool = True
    """Whether support multi-agent conversations"""

    support_vision: bool = True
    """Whether support vision data"""

    supported_blocks: list[type] = [
        TextBlock,
        # Multimodal
        ImageBlock,
        # Tool use
        ToolUseBlock,
        ToolResultBlock,
    ]
    """The list of supported message blocks"""

    def __init__(
        self,
        conversation_history_prompt: str = (
            "# Conversation History\n"
            "The content between <history></history> tags contains "
            "your conversation history\n"
        ),
        token_counter: TokenCounterBase | None = None,
        max_tokens: int | None = None,
    ) -> None:
        """Initialize the DashScope multi-agent formatter.

        Args:
            conversation_history_prompt (`str`):
                The prompt to use for the conversation history section.
        """
        super().__init__(token_counter=token_counter, max_tokens=max_tokens)
        self.conversation_history_prompt = conversation_history_prompt

    async def _format_tool_sequence(
        self,
        msgs: list[Msg],
    ) -> list[dict[str, Any]]:
        """Given a sequence of tool call/result messages, format them into
        the required format for the Anthropic API."""
        return await AnthropicChatFormatter().format(msgs)

    async def _format_agent_message(
        self,
        msgs: list[Msg],
        is_first: bool = True,
    ) -> list[dict[str, Any]]:
        """Given a sequence of messages without tool calls/results, format
        them into the required format for the Anthropic API."""

        if is_first:
            conversation_history_prompt = self.conversation_history_prompt
        else:
            conversation_history_prompt = ""

        # Format into required Anthropic format
        formatted_msgs: list[dict] = []

        # Collect the multimodal files
        conversation_blocks: list = []
        accumulated_text = []
        for msg in msgs:
            for block in msg.get_content_blocks():
                if block["type"] == "text":
                    accumulated_text.append(f"{msg.name}: {block['text']}")

                elif block["type"] == "image":
                    # Handle the accumulated text as a single block
                    if accumulated_text:
                        conversation_blocks.append(
                            {
                                "text": "\n".join(accumulated_text),
                                "type": "text",
                            },
                        )
                        accumulated_text.clear()

                    conversation_blocks.append({**block})

        if accumulated_text:
            conversation_blocks.append(
                {
                    "text": "\n".join(accumulated_text),
                    "type": "text",
                },
            )

        if conversation_blocks:
            if conversation_blocks[0].get("text"):
                conversation_blocks[0]["text"] = (
                    conversation_history_prompt
                    + "<history>\n"
                    + conversation_blocks[0]["text"]
                )

            else:
                conversation_blocks.insert(
                    0,
                    {
                        "type": "text",
                        "text": conversation_history_prompt + "<history>\n",
                    },
                )

            if conversation_blocks[-1].get("text"):
                conversation_blocks[-1]["text"] += "\n</history>"

            else:
                conversation_blocks.append(
                    {"type": "text", "text": "</history>"},
                )

        if conversation_blocks:
            formatted_msgs.append(
                {
                    "role": "user",
                    "content": conversation_blocks,
                },
            )

        return formatted_msgs



# -*- coding: utf-8 -*-
"""The message class in agentscope."""
from datetime import datetime
from typing import Literal, List, overload, Sequence

import shortuuid

from . import ToolResultBlock
from ._message_block import (
    TextBlock,
    ToolUseBlock,
    ImageBlock,
    AudioBlock,
    ContentBlock,
    VideoBlock,
    ThinkingBlock,
)
from ..types import JSONSerializableObject


class Msg:
    """The message class in agentscope."""

    def __init__(
        self,
        name: str,
        content: str | Sequence[ContentBlock],
        role: Literal["user", "assistant", "system"],
        metadata: dict[str, JSONSerializableObject] | None = None,
        timestamp: str | None = None,
        invocation_id: str | None = None,
    ) -> None:
        """Initialize the Msg object.

        Args:
            name (`str`):
                The name of the message sender.
            content (`str | list[ContentBlock]`):
                The content of the message.
            role (`Literal["user", "assistant", "system"]`):
                The role of the message sender.
            metadata (`dict[str, JSONSerializableObject] | None`, optional):
                The metadata of the message, e.g. structured output.
            timestamp (`str | None`, optional):
                The created timestamp of the message. If not given, the
                timestamp will be set automatically.
            invocation_id (`str | None`, optional):
                The related API invocation id, if any. This is useful for
                tracking the message in the context of an API call.
        """

        self.name = name

        assert isinstance(
            content,
            (list, str),
        ), "The content must be a string or a list of content blocks."

        self.content = content

        assert role in ["user", "assistant", "system"]
        self.role = role

        self.metadata = metadata

        self.id = shortuuid.uuid()
        self.timestamp = (
            timestamp
            or datetime.now().strftime(
                "%Y-%m-%d %H:%M:%S.%f",
            )[:-3]
        )
        self.invocation_id = invocation_id

    def to_dict(self) -> dict:
        """Convert the message into JSON dict data."""
        return {
            "id": self.id,
            "name": self.name,
            "role": self.role,
            "content": self.content,
            "metadata": self.metadata,
            "timestamp": self.timestamp,
        }

    @classmethod
    def from_dict(cls, json_data: dict) -> "Msg":
        """Load a message object from the given JSON data."""
        new_obj = cls(
            name=json_data["name"],
            content=json_data["content"],
            role=json_data["role"],
            metadata=json_data.get("metadata", None),
            timestamp=json_data.get("timestamp", None),
            invocation_id=json_data.get("invocation_id", None),
        )

        new_obj.id = json_data.get("id", new_obj.id)
        return new_obj

    def has_content_blocks(
        self,
        block_type: Literal[
            "text",
            "tool_use",
            "tool_result",
            "image",
            "audio",
            "video",
        ]
        | None = None,
    ) -> bool:
        """Check if the message has content blocks of the given type.

        Args:
            block_type (Literal["text", "tool_use", "tool_result", "image", \
            "audio", "video"] | None, defaults to None):
                The type of the block to be checked. If `None`, it will
                check if there are any content blocks.
        """
        return len(self.get_content_blocks(block_type)) > 0

    def get_text_content(self) -> str | None:
        """Get the pure text blocks from the message content."""
        if isinstance(self.content, str):
            return self.content

        gathered_text = None
        for block in self.content:
            if block.get("type") == "text":
                if gathered_text is None:
                    gathered_text = str(block.get("text"))
                else:
                    gathered_text += block.get("text")
        return gathered_text

    @overload
    def get_content_blocks(
        self,
        block_type: Literal["text"],
    ) -> List[TextBlock]:
        ...

    @overload
    def get_content_blocks(
        self,
        block_type: Literal["tool_use"],
    ) -> List[ToolUseBlock]:
        ...

    @overload
    def get_content_blocks(
        self,
        block_type: Literal["tool_result"],
    ) -> List[ToolResultBlock]:
        ...

    @overload
    def get_content_blocks(
        self,
        block_type: Literal["image"],
    ) -> List[ImageBlock]:
        ...

    @overload
    def get_content_blocks(
        self,
        block_type: Literal["audio"],
    ) -> List[AudioBlock]:
        ...

    @overload
    def get_content_blocks(
        self,
        block_type: Literal["video"],
    ) -> List[VideoBlock]:
        ...

    @overload
    def get_content_blocks(
        self,
        block_type: None = None,
    ) -> List[ContentBlock]:
        ...

    def get_content_blocks(
        self,
        block_type: Literal[
            "text",
            "thinking",
            "tool_use",
            "tool_result",
            "image",
            "audio",
            "video",
        ]
        | None = None,
    ) -> (
        List[ContentBlock]
        | List[TextBlock]
        | List[ThinkingBlock]
        | List[ToolUseBlock]
        | List[ToolResultBlock]
        | List[ImageBlock]
        | List[AudioBlock]
        | List[VideoBlock]
    ):
        """Get the content in block format. If the content is a string,
        it will be converted to a text block.

        Args:
            block_type (`Literal["text", "thinking", "tool_use", \
            "tool_result", "image", "audio", "video"] | None`, optional):
                The type of the block to be extracted. If `None`, all blocks
                will be returned.

        Returns:
            `List[ContentBlock]`:
                The content blocks.
        """
        blocks = []
        if isinstance(self.content, str):
            blocks.append(
                TextBlock(type="text", text=self.content),
            )
        else:
            blocks = self.content or []

        if block_type is not None:
            blocks = [_ for _ in blocks if _["type"] == block_type]

        return blocks

    def __repr__(self) -> str:
        """Get the string representation of the message."""
        return (
            f"Msg(id='{self.id}', "
            f"name='{self.name}', "
            f"content={repr(self.content)}, "
            f"role='{self.role}', "
            f"metadata={repr(self.metadata)}, "
            f"timestamp='{self.timestamp}', "
            f"invocation_id='{self.invocation_id}')"
        )



# -*- coding: utf-8 -*-
"""The toolkit class for tool calls in agentscope."""

import asyncio
import inspect
from copy import deepcopy
from dataclasses import dataclass
from functools import partial
from typing import (
    AsyncGenerator,
    Literal,
    Dict,
    Any,
    Type,
    Generator,
    Callable,
    Awaitable,
)

from pydantic import (
    BaseModel,
    Field,
    create_model,
    ConfigDict,
)
from docstring_parser import parse

from ._async_wrapper import (
    _async_generator_wrapper,
    _object_wrapper,
    _sync_generator_wrapper,
)
from ._registered_tool_function import RegisteredToolFunction
from ._response import ToolResponse
from .._utils._common import _remove_title_field
from ..mcp import (
    MCPToolFunction,
    MCPClientBase,
    StatefulClientBase,
)
from ..message import (
    ToolUseBlock,
    TextBlock,
)
from ..module import StateModule
from ..types import (
    JSONSerializableObject,
    ToolFunction,
)
from ..tracing._trace import trace_toolkit
from .._logging import logger


@dataclass
class ToolGroup:
    """The tool group class"""

    name: str
    """The group name, which will be used in the reset function as the group
    identifier."""
    active: bool
    """If the tool group is active, meaning the tool functions in this group
    is included in the JSON schema"""
    description: str
    """The description of the tool group to tell the agent what the tool
    group is about."""
    notes: str | None = None
    """The using notes of the tool group, to remind the agent how to use"""


class Toolkit(StateModule):
    """The class that supports both function- and group-level tool management.

    Use the following methods to manage the tool functions:

    - `register_tool_function`
    - `remove_tool_function`

    For group-level management:

    - `create_tool_group`
    - `update_tool_groups`
    - `remove_tool_groups`

    MCP related methods:

    - `register_mcp_server`
    - `remove_mcp_servers`

    To run the tool functions or get the data from the activated tools:

    - `call_tool_function`
    - `get_json_schemas`
    - `get_tool_group_notes`
    """

    def __init__(self) -> None:
        """Initialize the toolkit."""
        super().__init__()

        self.tools: dict[str, RegisteredToolFunction] = {}
        self.groups: dict[str, ToolGroup] = {}

    def create_tool_group(
        self,
        group_name: str,
        description: str,
        active: bool = False,
        notes: str | None = None,
    ) -> None:
        """Create a tool group to organize tool functions

        Args:
            group_name (`str`):
                The name of the tool group.
            description (`str`):
                The description of the tool group.
            active (`bool`, defaults to `False`):
                If the group is active, meaning the tool functions in this
                group are included in the JSON schema.
            notes (`str | None`, optional):
                The notes used to remind the agent how to use the tool
                functions properly, which can be combined into the system
                prompt.
        """
        if group_name in self.groups or group_name == "basic":
            raise ValueError(
                f"Tool group '{group_name}' is already registered in the "
                "toolkit.",
            )

        self.groups[group_name] = ToolGroup(
            name=group_name,
            description=description,
            notes=notes,
            active=active,
        )

    def update_tool_groups(self, group_names: list[str], active: bool) -> None:
        """Update the activation status of the given tool groups.

        Args:
            group_names (`list[str]`):
                The list of tool group names to be updated.
            active (`bool`):
                If the tool groups should be activated or deactivated.
        """

        for group_name in group_names:
            if group_name == "basic":
                logger.warning(
                    "The 'basic' tool group is always active, skipping it.",
                )

            if group_name in self.groups:
                self.groups[group_name].active = active

    def remove_tool_groups(self, group_names: str | list[str]) -> None:
        """Remove tool functions from the toolkit by their group names.

        Args:
            group_names (`str | list[str]`):
                The group names to be removed from the toolkit.
        """
        if isinstance(group_names, str):
            group_names = [group_names]

        if not isinstance(group_names, list) or not all(
            isinstance(_, str) for _ in group_names
        ):
            raise TypeError(
                f"The group_names must be a list of strings, "
                f"but got {type(group_names)}.",
            )

        if "basic" in group_names:
            raise ValueError(
                "Cannot remove the default 'basic' tool group.",
            )

        for group_name in group_names:
            self.groups.pop(group_name, None)

        # Remove the tool functions in the given groups
        tool_names = deepcopy(list(self.tools.keys()))
        for tool_name in tool_names:
            if self.tools[tool_name].group in group_names:
                self.tools.pop(tool_name)

    def register_tool_function(  # pylint: disable=too-many-branches
        self,
        tool_func: ToolFunction,
        group_name: str | Literal["basic"] = "basic",
        preset_kwargs: dict[str, JSONSerializableObject] | None = None,
        func_description: str | None = None,
        json_schema: dict | None = None,
        include_long_description: bool = True,
        include_var_positional: bool = False,
        include_var_keyword: bool = False,
        postprocess_func: (
            Callable[
                [ToolUseBlock, ToolResponse],
                ToolResponse | None,
            ]
            | Callable[
                [ToolUseBlock, ToolResponse],
                Awaitable[ToolResponse | None],
            ]
        )
        | None = None,
    ) -> None:
        """Register a tool function to the toolkit.

        Args:
            tool_func (`ToolFunction`):
                The tool function, which can be async or sync, streaming or
                not-streaming, but the response must be a `ToolResponse`
                object.
            group_name (`str | Literal["basic"]`, defaults to `"basic"`):
                The belonging group of the tool function. Tools in "basic"
                group is always included in the JSON schema, while the others
                are only included when their group is active.
            preset_kwargs (`dict[str, JSONSerializableObject] | None`, \
            optional):
                Preset arguments by the user, which will not be included in
                the JSON schema, nor exposed to the agent.
            func_description (`str | None`, optional):
                The function description. If not provided, the description
                will be extracted from the docstring automatically.
            json_schema (`dict | None`, optional):
                Manually provided JSON schema for the tool function, which
                should be `{"type": "function", "function": {"name":
                "function_name": "xx", "description": "xx",
                "parameters": {...}}}`
            include_long_description (`bool`, defaults to `True`):
                When extracting function description from the docstring, if
                the long description will be included.
            include_var_positional (`bool`, defaults to `False`):
                Whether to include the variable positional arguments (`*args`)
                in the function schema.
            include_var_keyword (`bool`, defaults to `False`):
                Whether to include the variable keyword arguments (`**kwargs`)
                in the function schema.
            postprocess_func (`(Callable[[ToolUseBlock, ToolResponse], \
            ToolResponse | None] | Callable[[ToolUseBlock, ToolResponse], \
            Awaitable[ToolResponse | None]]) | None`, optional):
                A post-processing function that will be called after the tool
                function is executed, taking the tool call block and tool
                response as arguments. The function can be either sync or
                async. If it returns `None`, the tool result will be
                returned as is. If it returns a `ToolResponse`,
                the returned block will be used as the final tool result.
        """
        # Arguments checking
        if group_name not in self.groups and group_name != "basic":
            raise ValueError(
                f"Tool group '{group_name}' not found.",
            )

        # Check the manually provided JSON schema if provided
        if json_schema:
            assert (
                isinstance(json_schema, dict)
                and "type" in json_schema
                and json_schema["type"] == "function"
                and "function" in json_schema
                and isinstance(json_schema["function"], dict)
            ), "Invalid JSON schema for the tool function."

        # Handle MCP tool function and regular function respectively
        mcp_name = None
        if isinstance(tool_func, MCPToolFunction):
            func_name = tool_func.name
            original_func = tool_func.__call__
            self._validate_tool_function(func_name)
            json_schema = json_schema or tool_func.json_schema
            mcp_name = tool_func.mcp_name

        elif isinstance(tool_func, partial):
            # partial function
            kwargs = tool_func.keywords
            # Turn args into keyword arguments
            if tool_func.args:
                param_names = list(
                    inspect.signature(tool_func.func).parameters.keys(),
                )
                for i, arg in enumerate(tool_func.args):
                    if i < len(param_names):
                        kwargs[param_names[i]] = arg

            preset_kwargs = {
                **kwargs,
                **(preset_kwargs or {}),
            }

            func_name = tool_func.func.__name__
            original_func = tool_func.func
            self._validate_tool_function(func_name)
            json_schema = json_schema or self._parse_tool_function(
                tool_func.func,
                include_long_description=include_long_description,
                include_var_positional=include_var_positional,
                include_var_keyword=include_var_keyword,
            )

        else:
            # normal function
            func_name = tool_func.__name__
            original_func = tool_func
            self._validate_tool_function(func_name)
            json_schema = json_schema or self._parse_tool_function(
                tool_func,
                include_long_description=include_long_description,
                include_var_positional=include_var_positional,
                include_var_keyword=include_var_keyword,
            )

        # Override the description if provided
        if func_description:
            json_schema["function"]["description"] = func_description

        # Remove the preset kwargs from the JSON schema
        for arg_name in preset_kwargs or {}:
            if arg_name in json_schema["function"]["parameters"]["properties"]:
                json_schema["function"]["parameters"]["properties"].pop(
                    arg_name,
                )

        if "required" in json_schema["function"]["parameters"]:
            for arg_name in preset_kwargs or {}:
                if (
                    arg_name
                    in json_schema["function"]["parameters"]["required"]
                ):
                    json_schema["function"]["parameters"]["required"].remove(
                        arg_name,
                    )

            # Remove the required field if it is empty
            if len(json_schema["function"]["parameters"]["required"]) == 0:
                json_schema["function"]["parameters"].pop("required", None)

        func_obj = RegisteredToolFunction(
            name=func_name,
            group=group_name,
            source="function",
            original_func=original_func,
            json_schema=json_schema,
            preset_kwargs=preset_kwargs or {},
            extended_model=None,
            mcp_name=mcp_name,
            postprocess_func=postprocess_func,
        )

        self.tools[func_name] = func_obj

    def remove_tool_function(self, tool_name: str) -> None:
        """Remove tool function from the toolkit by its name.

        Args:
            tool_name (`str`):
                The name of the tool function to be removed.
        """

        if tool_name not in self.tools:
            logger.warning(
                "Skipping removing tool function '%s' as it does not exist.",
                tool_name,
            )

        self.tools.pop(tool_name, None)

    def get_json_schemas(
        self,
    ) -> list[dict]:
        """Get the JSON schemas from the tool functions that belong to the
        active groups.

        .. note:: The preset keyword arguments is removed from the JSON
         schema, and the extended model is applied if it is set.

        Example:
            .. code-block:: JSON
                :caption: Example of tool function JSON schemas

                [
                    {
                        "type": "function",
                        "function": {
                            "name": "google_search",
                            "description": "Search on Google.",
                            "parameters": {
                                "type": "object",
                                "properties": {
                                    "query": {
                                        "type": "string",
                                        "description": "The search query."
                                    }
                                },
                                "required": ["query"]
                            }
                        }
                    },
                    ...
                ]

        Returns:
            `list[dict]`:
                A list of function JSON schemas.
        """
        # If meta tool is set here, update its extended model here
        if "reset_equipped_tools" in self.tools:
            fields = {}
            for group_name, group in self.groups.items():
                if group_name == "basic":
                    continue
                fields[group_name] = (
                    bool,
                    Field(
                        default=False,
                        description=group.description,
                    ),
                )
            extended_model = create_model("_DynamicModel", **fields)
            self.set_extended_model(
                "reset_equipped_tools",
                extended_model,
            )

        return [
            tool.extended_json_schema
            for tool in self.tools.values()
            if tool.group == "basic" or self.groups[tool.group].active
        ]

    def set_extended_model(
        self,
        func_name: str,
        model: Type[BaseModel] | None,
    ) -> None:
        """Set the extended model for a tool function, so that the original
        JSON schema will be extended.

        Args:
            func_name (`str`):
                The name of the tool function.
            model (`Union[Type[BaseModel], None]`):
                The extended model to be set.
        """
        if model is not None and not issubclass(model, BaseModel):
            raise TypeError(
                "The extended model must be a child class of pydantic "
                f"BaseModel, but got {type(model)}.",
            )

        if func_name in self.tools:
            self.tools[func_name].extended_model = model

        else:
            raise ValueError(
                f"Tool function '{func_name}' not found in the toolkit.",
            )

    async def remove_mcp_clients(
        self,
        client_names: list[str],
    ) -> None:
        """Remove tool functions from the MCP clients by their names.

        Args:
            client_names (`list[str]`):
                The names of the MCP client, which used to initialize the
                client instance.
        """
        if isinstance(client_names, str):
            client_names = [client_names]

        if isinstance(client_names, list) and not all(
            isinstance(_, str) for _ in client_names
        ):
            raise TypeError(
                f"The client_names must be a list of strings, "
                f"but got {type(client_names)}.",
            )

        to_removed = []
        func_names = deepcopy(list(self.tools.keys()))
        for func_name in func_names:
            if self.tools[func_name].mcp_name in client_names:
                self.tools.pop(func_name)
                to_removed.append(func_name)

        logger.info(
            "Removed %d tool functions from %d MCP: %s",
            len(to_removed),
            len(client_names),
            ", ".join(to_removed),
        )

    @trace_toolkit
    async def call_tool_function(
        self,
        tool_call: ToolUseBlock,
    ) -> AsyncGenerator[ToolResponse, None]:
        """Execute the tool function by the `ToolUseBlock` and return the
        tool response chunk in unified streaming mode, i.e. an async
        generator of `ToolResponse` objects.

        .. note:: The tool response chunk is **accumulated**.

        Args:
            tool_call (`ToolUseBlock`):
                A tool call block.

        Yields:
            `ToolResponse`:
                The tool response chunk, in accumulative manner.
        """

        # Check
        if tool_call["name"] not in self.tools:
            return _object_wrapper(
                ToolResponse(
                    content=[
                        TextBlock(
                            type="text",
                            text="FunctionNotFoundError: Cannot find the "
                            f"function named {tool_call['name']}",
                        ),
                    ],
                ),
                None,
            )

        # Prepare function and keyword arguments
        tool_func = self.tools[tool_call["name"]]
        kwargs = {
            **tool_func.preset_kwargs,
            **(tool_call.get("input", {}) or {}),
        }

        # Prepare postprocess function
        if tool_func.postprocess_func:
            # Type: partial wraps the postprocess_func with tool_call bound,
            # reducing it from (ToolUseBlock, ToolResponse) to (ToolResponse)
            partial_postprocess_func: (
                Callable[[ToolResponse], ToolResponse | None]
                | Callable[[ToolResponse], Awaitable[ToolResponse | None]]
            ) | None = partial(
                tool_func.postprocess_func,
                tool_call,
            )
        else:
            partial_postprocess_func = None

        # Async function
        try:
            if inspect.iscoroutinefunction(tool_func.original_func):
                try:
                    res = await tool_func.original_func(**kwargs)
                except asyncio.CancelledError:
                    res = ToolResponse(
                        content=[
                            TextBlock(
                                type="text",
                                text="<system-info>"
                                "The tool call has been interrupted "
                                "by the user."
                                "</system-info>",
                            ),
                        ],
                        stream=True,
                        is_last=True,
                        is_interrupted=True,
                    )

            else:
                # When `tool_func.original_func` is Async generator function or
                # Sync function
                res = tool_func.original_func(**kwargs)

        except Exception as e:
            res = ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"Error: {e}",
                    ),
                ],
            )

        # Handle different return type

        # If return an async generator
        if isinstance(res, AsyncGenerator):
            return _async_generator_wrapper(res, partial_postprocess_func)

        # If return a sync generator
        if isinstance(res, Generator):
            return _sync_generator_wrapper(res, partial_postprocess_func)

        if isinstance(res, ToolResponse):
            return _object_wrapper(res, partial_postprocess_func)

        raise TypeError(
            "The tool function must return a ToolResponse object, or an "
            "AsyncGenerator/Generator of ToolResponse objects, "
            f"but got {type(res)}.",
        )

    async def register_mcp_client(
        self,
        mcp_client: MCPClientBase,
        group_name: str = "basic",
        enable_funcs: list[str] | None = None,
        disable_funcs: list[str] | None = None,
        preset_kwargs_mapping: dict[str, dict[str, Any]] | None = None,
        postprocess_func: (
            Callable[
                [ToolUseBlock, ToolResponse],
                ToolResponse | None,
            ]
            | Callable[
                [ToolUseBlock, ToolResponse],
                Awaitable[ToolResponse | None],
            ]
        )
        | None = None,
    ) -> None:
        """Register tool functions from an MCP client.

        Args:
            mcp_client (`MCPClientBase`):
                The MCP client instance to connect to the MCP server.
            group_name (`str`, defaults to `"basic"`):
                The group name that the tool functions will be added to.
            enable_funcs (`list[str] | None`, optional):
                The functions to be added into the toolkit. If `None`, all
                tool functions within the MCP servers will be added.
            disable_funcs (`list[str] | None`, optional):
                The functions that will be filtered out. If `None`, no
                tool functions will be filtered out.
            preset_kwargs_mapping: (`Optional[dict[str, dict[str, Any]]]`, \
            defaults to `None`):
                The preset keyword arguments mapping, whose keys are the tool
                function names and values are the preset keyword arguments.
            postprocess_func (`(Callable[[ToolUseBlock, ToolResponse], \
            ToolResponse | None] | Callable[[ToolUseBlock, ToolResponse], \
            Awaitable[ToolResponse | None]]) | None`, optional):
                A post-processing function that will be called after the tool
                function is executed, taking the tool call block and tool
                response as arguments. The function can be either sync or
                async. If it returns `None`, the tool result will be
                returned as is. If it returns a `ToolResponse`,
                the returned block will be used as the final tool result.
        """
        if (
            isinstance(mcp_client, StatefulClientBase)
            and not mcp_client.is_connected
        ):
            raise RuntimeError(
                "The MCP client is not connected to the server. Use the "
                "`connect()` method first.",
            )

        # Check arguments for enable_funcs and disabled_funcs
        if enable_funcs is not None and disable_funcs is not None:
            assert isinstance(enable_funcs, list) and all(
                isinstance(_, str) for _ in enable_funcs
            ), (
                "Enable functions should be a list of strings, but got "
                f"{enable_funcs}."
            )

            assert isinstance(disable_funcs, list) and all(
                isinstance(_, str) for _ in disable_funcs
            ), (
                "Disable functions should be a list of strings, but got "
                f"{disable_funcs}."
            )
            intersection = set(enable_funcs).intersection(
                set(disable_funcs),
            )
            assert len(intersection) == 0, (
                f"The functions in enable_funcs and disable_funcs "
                f"should not overlap, but got {intersection}."
            )

        if not (
            preset_kwargs_mapping is None
            or isinstance(preset_kwargs_mapping, dict)
        ):
            raise TypeError(
                f"The preset_kwargs_mapping must be a dictionary or None, "
                f"but got {type(preset_kwargs_mapping)}.",
            )

        tool_names = []
        for mcp_tool in await mcp_client.list_tools():
            # Skip the functions that are not in the enable_funcs if
            # enable_funcs is not None
            if enable_funcs is not None and mcp_tool.name not in enable_funcs:
                continue

            # Skip the disabled functions
            if disable_funcs is not None and mcp_tool.name in disable_funcs:
                continue

            tool_names.append(mcp_tool.name)

            # Obtain callable function object
            func_obj = await mcp_client.get_callable_function(
                func_name=mcp_tool.name,
                wrap_tool_result=True,
            )

            # Prepare preset kwargs
            preset_kwargs = None
            if preset_kwargs_mapping is not None:
                preset_kwargs = preset_kwargs_mapping.get(mcp_tool.name, {})

            # TODO: handle mcp_server_name
            self.register_tool_function(
                tool_func=func_obj,
                group_name=group_name,
                preset_kwargs=preset_kwargs,
                postprocess_func=postprocess_func,
            )

        logger.info(
            "Registered %d tool functions from MCP: %s.",
            len(tool_names),
            ", ".join(tool_names),
        )

    def state_dict(self) -> dict[str, Any]:
        """Get the state dictionary of the toolkit.

        Returns:
            `dict[str, Any]`:
                A dictionary containing the active tool group names.
        """
        return {
            "active_groups": [
                name for name, group in self.groups.items() if group.active
            ],
        }

    def load_state_dict(
        self,
        state_dict: dict[str, Any],
        strict: bool = True,
    ) -> None:
        """Load the state dictionary into the toolkit.

        Args:
            state_dict (`dict`):
                The state dictionary to load, which should have "active_groups"
                key and its value must be a list of group names.
            strict (`bool`, defaults to `True`):
                If `True`, raises an error if any key in the module is not
                found in the state_dict. If `False`, skips missing keys.
        """
        if (
            not isinstance(state_dict, dict)
            or "active_groups" not in state_dict
            or not isinstance(state_dict["active_groups"], list)
        ):
            raise ValueError(
                "The state_dict for toolkit must be a dictionary with "
                "active_groups key and its value must be a list, "
                f"but got {type(state_dict)}.",
            )

        if strict and list(state_dict.keys()) != ["active_groups"]:
            raise ValueError(
                "Get additional keys in the state_dict: "
                f'{list(state_dict.keys())}, but only "active_groups" '
                "is expected.",
            )

        for group_name, group in self.groups.items():
            if group_name in state_dict["active_groups"]:
                group.active = True
            else:
                group.active = False

    def get_activated_notes(self) -> str:
        """Get the notes from the active tool groups, which can be used to
        construct the system prompt for the agent.

        Returns:
            `str`:
                The combined notes from the active tool groups.
        """
        collected_notes = []
        for group_name, group in self.groups.items():
            if group.active and group.notes:
                collected_notes.append(
                    "\n".join(
                        [f"## About {group_name} Tools", group.notes],
                    ),
                )
        return "\n".join(collected_notes)

    def reset_equipped_tools(self, **kwargs: Any) -> ToolResponse:
        """Choose appropriate tools to equip yourself with, so that you can
        finish your task. Each argument in this function represents a group
        of related tools, and the value indicates whether to activate the
        group or not. Besides, the tool response of this function will
        contain the precaution notes for using them, which you
        **MUST pay attention to and follow**. You can also reuse this function
        to check the notes of the tool groups.

        Note this function will `reset` the tools, so that the original tools
        will be removed first.
        """

        to_activate = []
        for key, value in kwargs.items():
            if not isinstance(value, bool):
                return ToolResponse(
                    content=[
                        TextBlock(
                            type="text",
                            text=f"Invalid arguments: the argument {key} "
                            f"should be a bool value, but got {type(value)}.",
                        ),
                    ],
                )

            if value:
                to_activate.append(key)

        self.update_tool_groups(to_activate, active=True)

        notes = self.get_activated_notes()

        return ToolResponse(
            content=[
                TextBlock(
                    type="text",
                    text=f"Active tool groups successfully: {to_activate}. "
                    "You MUST follow these notes to use the tools:\n"
                    f"<notes>{notes}</notes>",
                ),
            ],
        )

    def clear(self) -> None:
        """Clear the toolkit, removing all tool functions and groups."""
        self.tools.clear()
        self.groups.clear()

    def _validate_tool_function(self, func_name: str) -> None:
        """Check if the tool function already registered in the toolkit. If
        so, raise a ValueError."""
        if func_name in self.tools:
            raise ValueError(
                f"A function with name '{func_name} is already registered "
                "in the toolkit.",
            )

    @staticmethod
    def _parse_tool_function(
        tool_func: ToolFunction,
        include_long_description: bool,
        include_var_positional: bool,
        include_var_keyword: bool,
    ) -> dict:
        """Extract JSON schema from the tool function's docstring"""
        docstring = parse(tool_func.__doc__)
        params_docstring = {
            _.arg_name: _.description for _ in docstring.params
        }

        # Function description
        descriptions = []
        if docstring.short_description is not None:
            descriptions.append(docstring.short_description)

        if include_long_description and docstring.long_description is not None:
            descriptions.append(docstring.long_description)

        func_description = "\n\n".join(descriptions)

        # Create a dynamic model with the function signature
        fields = {}
        for name, param in inspect.signature(tool_func).parameters.items():
            # Skip the `self` and `cls` parameters
            if name in ["self", "cls"]:
                continue

            # Handle `**kwargs`
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                if not include_var_keyword:
                    continue

                fields[name] = (
                    Dict[str, Any]
                    if param.annotation == inspect.Parameter.empty
                    else Dict[str, param.annotation],  # type: ignore
                    Field(
                        description=params_docstring.get(
                            f"**{name}",
                            params_docstring.get(name, None),
                        ),
                        default={}
                        if param.default is param.empty
                        else param.default,
                    ),
                )

            elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                if not include_var_positional:
                    continue

                fields[name] = (
                    list[Any]
                    if param.annotation == inspect.Parameter.empty
                    else list[param.annotation],  # type: ignore
                    Field(
                        description=params_docstring.get(
                            f"*{name}",
                            params_docstring.get(name, None),
                        ),
                        default=[]
                        if param.default is param.empty
                        else param.default,
                    ),
                )

            else:
                fields[name] = (
                    Any
                    if param.annotation == inspect.Parameter.empty
                    else param.annotation,
                    Field(
                        description=params_docstring.get(name, None),
                        default=...
                        if param.default is param.empty
                        else param.default,
                    ),
                )

        base_model = create_model(
            "_StructuredOutputDynamicClass",
            __config__=ConfigDict(arbitrary_types_allowed=True),
            **fields,
        )
        params_json_schema = base_model.model_json_schema()

        # Remove the title from the json schema
        _remove_title_field(params_json_schema)

        func_json_schema: dict = {
            "type": "function",
            "function": {
                "name": tool_func.__name__,
                "parameters": params_json_schema,
            },
        }

        if func_description not in [None, ""]:
            func_json_schema["function"]["description"] = func_description

        return func_json_schema



# -*- coding: utf-8 -*-
"""The text reader that reads text into vector records."""
import hashlib
import os
from typing import Literal

from ._reader_base import ReaderBase, Document
from .._document import DocMetadata
from ..._logging import logger
from ...message import TextBlock


class TextReader(ReaderBase):
    """The text reader that splits text into chunks by a fixed chunk size
    and chunk overlap."""

    def __init__(
        self,
        chunk_size: int = 512,
        split_by: Literal["char", "sentence", "paragraph"] = "sentence",
    ) -> None:
        """Initialize the text reader.

        Args:
            chunk_size (`int`, default to 512):
                The size of each chunk, in number of characters.
            split_by (`Literal["char", "paragraph"]`, default to \
            "sentence"):
                The unit to split the text, can be "char", "sentence", or
                "paragraph". Note that "sentence" is implemented by "nltk"
                library, which only supports English text.
        """
        if chunk_size <= 0:
            raise ValueError(
                f"The chunk_size must be positive, got {chunk_size}",
            )

        if split_by not in ["char", "sentence", "paragraph"]:
            raise ValueError(
                "The split_by must be one of 'char', 'sentence' or "
                f"'paragraph', got {split_by}",
            )

        self.chunk_size = chunk_size
        self.split_by = split_by

    async def __call__(
        self,
        text: str,
    ) -> list[Document]:
        """Read a text string, split it into chunks, and return a list of
        Document objects.

        Args:
            text (`str`):
                The input text string, or a path to the local text file.

        Returns:
            `list[Document]`:
                A list of Document objects, where the metadata contains the
                chunked text, doc id and chunk id.
        """
        if os.path.exists(text) and os.path.isfile(text):
            logger.info("Reading text from local file: %s", text)
            with open(text, "r", encoding="utf-8") as file:
                text = file.read()

        logger.info(
            "Reading text with chunk_size=%d, split_by=%s",
            self.chunk_size,
            self.split_by,
        )
        splits = []
        if self.split_by == "char":
            # Split by character
            for i in range(0, len(text), self.chunk_size):
                start = max(0, i)
                end = min(i + self.chunk_size, len(text))
                splits.append(text[start:end])

        elif self.split_by == "sentence":
            try:
                import nltk

                nltk.download("punkt", quiet=True)
                nltk.download("punkt_tab", quiet=True)
            except ImportError as e:
                raise ImportError(
                    "nltk is not installed. Please install it with "
                    "`pip install nltk`.",
                ) from e

            sentences = nltk.sent_tokenize(text)

            # Handle the chunk_size for sentences
            processed_sentences = []
            for _ in sentences:
                if len(_) <= self.chunk_size:
                    processed_sentences.append(_)
                else:
                    # If the sentence itself exceeds chunk size, we need to
                    # truncate it
                    chunks = [
                        _[j : j + self.chunk_size]
                        for j in range(0, len(_), self.chunk_size)
                    ]
                    processed_sentences.extend(chunks)

            splits.extend(processed_sentences)

        elif self.split_by == "paragraph":
            paragraphs = [_ for _ in text.split("\n") if len(_)]
            for para in paragraphs:
                if len(para) <= self.chunk_size:
                    splits.append(para)

                else:
                    # If the paragraph itself exceeds chunk size, we need to
                    # truncate it
                    chunks = [
                        para[k : k + self.chunk_size]
                        for k in range(0, len(para), self.chunk_size)
                    ]
                    splits.extend(chunks)

        logger.info(
            "Finished splitting the text into %d chunks.",
            len(splits),
        )

        doc_id = self.get_doc_id(text)

        return [
            Document(
                id=doc_id,
                metadata=DocMetadata(
                    content=TextBlock(type="text", text=_),
                    doc_id=doc_id,
                    chunk_id=idx,
                    total_chunks=len(splits),
                ),
            )
            for idx, _ in enumerate(splits)
        ]

    def get_doc_id(self, text: str) -> str:
        """Get the document ID. This function can be used to check if the
        doc_id already exists in the knowledge base."""
        return hashlib.sha256(text.encode("utf-8")).hexdigest()



# -*- coding: utf-8 -*-
"""The functions that wrap object, sync generator, and async generator
into async generators.

TODO: handle the exception raised when yielding from async generator
 into a normal ToolResponse instance.
"""
import asyncio
from typing import AsyncGenerator, Generator, Callable, Awaitable

from ._response import ToolResponse
from ..message import TextBlock
from .._utils._common import _execute_async_or_sync_func


async def _postprocess_tool_response(
    tool_response: ToolResponse,
    postprocess_func: (
        Callable[[ToolResponse], ToolResponse | None]
        | Callable[[ToolResponse], Awaitable[ToolResponse | None]]
    )
    | None,
) -> ToolResponse:
    """Post-process a ToolResponse object with the given function.

    Supports both sync and async postprocess_func.
    """
    if postprocess_func:
        processed_response = await _execute_async_or_sync_func(
            postprocess_func,
            tool_response,
        )
        if processed_response:
            return processed_response
    return tool_response


async def _object_wrapper(
    obj: ToolResponse,
    postprocess_func: (
        Callable[[ToolResponse], ToolResponse | None]
        | Callable[[ToolResponse], Awaitable[ToolResponse | None]]
    )
    | None,
) -> AsyncGenerator[ToolResponse, None]:
    """Wrap a ToolResponse object to an async generator."""
    yield await _postprocess_tool_response(obj, postprocess_func)


async def _sync_generator_wrapper(
    sync_generator: Generator[ToolResponse, None, None],
    postprocess_func: (
        Callable[[ToolResponse], ToolResponse | None]
        | Callable[[ToolResponse], Awaitable[ToolResponse | None]]
    )
    | None,
) -> AsyncGenerator[ToolResponse, None]:
    """Wrap a sync generator to an async generator."""
    for chunk in sync_generator:
        yield await _postprocess_tool_response(chunk, postprocess_func)


async def _async_generator_wrapper(
    async_func: AsyncGenerator[ToolResponse, None],
    postprocess_func: (
        Callable[[ToolResponse], ToolResponse | None]
        | Callable[[ToolResponse], Awaitable[ToolResponse | None]]
    )
    | None,
) -> AsyncGenerator[ToolResponse, None]:
    """When the function is interrupted during generating the tool
    response, add an interrupted message to the response, and postpone
    the CancelledError to the caller."""

    last_chunk = None
    try:
        async for chunk in async_func:
            processed_chunk = await _postprocess_tool_response(
                chunk,
                postprocess_func,
            )
            yield processed_chunk
            last_chunk = processed_chunk

    except asyncio.CancelledError:
        interrupted_info = TextBlock(
            type="text",
            text="<system-info>"
            "The tool call has been interrupted by the user."
            "</system-info>",
        )
        if last_chunk:
            last_chunk.content.append(interrupted_info)
            last_chunk.is_interrupted = True
            last_chunk.is_last = True
            yield await _postprocess_tool_response(
                last_chunk,
                postprocess_func,
            )

        else:
            yield await _postprocess_tool_response(
                ToolResponse(
                    content=[interrupted_info],
                    is_interrupted=True,
                    is_last=True,
                ),
                postprocess_func,
            )



# -*- coding: utf-8 -*-
# pylint: disable=too-many-branches
"""The OpenAI formatter for agentscope."""
import base64
import json
import os
from typing import Any
from urllib.parse import urlparse

import requests

from ._truncated_formatter_base import TruncatedFormatterBase
from .._logging import logger
from ..message import (
    Msg,
    URLSource,
    TextBlock,
    ImageBlock,
    AudioBlock,
    Base64Source,
    ToolUseBlock,
    ToolResultBlock,
)
from ..token import TokenCounterBase


def _to_openai_image_url(url: str) -> str:
    """Convert an image url to openai format. If the given url is a local
    file, it will be converted to base64 format. Otherwise, it will be
    returned directly.

    Args:
        url (`str`):
            The local or public url of the image.
    """
    # See https://platform.openai.com/docs/guides/vision for details of
    # support image extensions.
    support_image_extensions = (
        ".png",
        ".jpg",
        ".jpeg",
        ".gif",
        ".webp",
    )

    parsed_url = urlparse(url)

    lower_url = url.lower()

    # Web url
    if not os.path.exists(url) and parsed_url.scheme != "":
        if any(lower_url.endswith(_) for _ in support_image_extensions):
            return url

    # Check if it is a local file
    elif os.path.exists(url) and os.path.isfile(url):
        if any(lower_url.endswith(_) for _ in support_image_extensions):
            with open(url, "rb") as image_file:
                base64_image = base64.b64encode(image_file.read()).decode(
                    "utf-8",
                )
            extension = parsed_url.path.lower().split(".")[-1]
            mime_type = f"image/{extension}"
            return f"data:{mime_type};base64,{base64_image}"

    raise TypeError(f'"{url}" should end with {support_image_extensions}.')


def _to_openai_audio_data(source: URLSource | Base64Source) -> dict:
    """Covert an audio source to OpenAI format."""
    if source["type"] == "url":
        extension = source["url"].split(".")[-1].lower()
        if extension not in ["wav", "mp3"]:
            raise TypeError(
                f"Unsupported audio file extension: {extension}, "
                "wav and mp3 are supported.",
            )

        parsed_url = urlparse(source["url"])

        if os.path.exists(source["url"]):
            with open(source["url"], "rb") as audio_file:
                data = base64.b64encode(audio_file.read()).decode("utf-8")

        # web url
        elif parsed_url.scheme != "":
            response = requests.get(source["url"])
            response.raise_for_status()
            data = base64.b64encode(response.content).decode("utf-8")

        else:
            raise ValueError(
                f"Unsupported audio source: {source['url']}, "
                "it should be a local file or a web URL.",
            )

        return {
            "data": data,
            "format": extension,
        }

    if source["type"] == "base64":
        data = source["data"]
        media_type = source["media_type"]

        if media_type not in ["audio/wav", "audio/mp3"]:
            raise TypeError(
                f"Unsupported audio media type: {media_type}, "
                "only audio/wav and audio/mp3 are supported.",
            )

        return {
            "data": data,
            "format": media_type.split("/")[-1],
        }

    raise TypeError(f"Unsupported audio source: {source['type']}.")


class OpenAIChatFormatter(TruncatedFormatterBase):
    """The class used to format message objects into the OpenAI API required
    format."""

    support_tools_api: bool = True
    """Whether support tools API"""

    support_multiagent: bool = True
    """Whether support multi-agent conversation"""

    support_vision: bool = True
    """Whether support vision models"""

    supported_blocks: list[type] = [
        TextBlock,
        ImageBlock,
        AudioBlock,
        ToolUseBlock,
        ToolResultBlock,
    ]
    """Supported message blocks for OpenAI API"""

    async def _format(
        self,
        msgs: list[Msg],
    ) -> list[dict[str, Any]]:
        """Format message objects into OpenAI API required format.

        Args:
            msgs (`list[Msg]`):
                The list of Msg objects to format.

        Returns:
            `list[dict[str, Any]]`:
                A list of dictionaries, where each dictionary has "name",
                "role", and "content" keys.
        """
        self.assert_list_of_msgs(msgs)

        messages: list[dict] = []
        for msg in msgs:
            content_blocks = []
            tool_calls = []

            for block in msg.get_content_blocks():
                typ = block.get("type")
                if typ == "text":
                    content_blocks.append({**block})

                elif typ == "tool_use":
                    tool_calls.append(
                        {
                            "id": block.get("id"),
                            "type": "function",
                            "function": {
                                "name": block.get("name"),
                                "arguments": json.dumps(
                                    block.get("input", {}),
                                    ensure_ascii=False,
                                ),
                            },
                        },
                    )

                elif typ == "tool_result":
                    messages.append(
                        {
                            "role": "tool",
                            "tool_call_id": block.get("id"),
                            "content": self.convert_tool_result_to_string(
                                block.get("output"),  # type: ignore[arg-type]
                            ),
                            "name": block.get("name"),
                        },
                    )

                elif typ == "image":
                    source_type = block["source"]["type"]
                    if source_type == "url":
                        url = _to_openai_image_url(block["source"]["url"])

                    elif source_type == "base64":
                        data = block["source"]["data"]
                        media_type = block["source"]["media_type"]
                        url = f"data:{media_type};base64,{data}"

                    else:
                        raise ValueError(
                            f"Unsupported image source type: {source_type}",
                        )

                    content_blocks.append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": url,
                            },
                        },
                    )

                elif typ == "audio":
                    input_audio = _to_openai_audio_data(block["source"])
                    content_blocks.append(
                        {
                            "type": "input_audio",
                            "input_audio": input_audio,
                        },
                    )

                else:
                    logger.warning(
                        "Unsupported block type %s in the message, skipped.",
                        typ,
                    )

            msg_openai = {
                "role": msg.role,
                "name": msg.name,
                "content": content_blocks or None,
            }

            if tool_calls:
                msg_openai["tool_calls"] = tool_calls

            # When both content and tool_calls are None, skipped
            if msg_openai["content"] or msg_openai.get("tool_calls"):
                messages.append(msg_openai)

        return messages


class OpenAIMultiAgentFormatter(TruncatedFormatterBase):
    """
    OpenAI formatter for multi-agent conversations, where more than
    a user and an agent are involved.
    .. tip:: This formatter is compatible with OpenAI API and
    OpenAI-compatible services like vLLM, Azure OpenAI, and others.
    """

    support_tools_api: bool = True
    """Whether support tools API"""

    support_multiagent: bool = True
    """Whether support multi-agent conversation"""

    support_vision: bool = True
    """Whether support vision models"""

    supported_blocks: list[type] = [
        TextBlock,
        ImageBlock,
        AudioBlock,
        ToolUseBlock,
        ToolResultBlock,
    ]
    """Supported message blocks for OpenAI API"""

    def __init__(
        self,
        conversation_history_prompt: str = (
            "# Conversation History\n"
            "The content between <history></history> tags contains "
            "your conversation history\n"
        ),
        token_counter: TokenCounterBase | None = None,
        max_tokens: int | None = None,
    ) -> None:
        """Initialize the OpenAI multi-agent formatter.

        Args:
            conversation_history_prompt (`str`):
                The prompt to use for the conversation history section.
        """
        super().__init__(token_counter=token_counter, max_tokens=max_tokens)
        self.conversation_history_prompt = conversation_history_prompt

    async def _format_tool_sequence(
        self,
        msgs: list[Msg],
    ) -> list[dict[str, Any]]:
        """Given a sequence of tool call/result messages, format them into
        the required format for the OpenAI API."""
        return await OpenAIChatFormatter().format(msgs)

    async def _format_agent_message(
        self,
        msgs: list[Msg],
        is_first: bool = True,
    ) -> list[dict[str, Any]]:
        """Given a sequence of messages without tool calls/results, format
        them into the required format for the OpenAI API."""

        if is_first:
            conversation_history_prompt = self.conversation_history_prompt
        else:
            conversation_history_prompt = ""

        # Format into required OpenAI format
        formatted_msgs: list[dict] = []

        conversation_blocks: list = []
        accumulated_text = []
        images = []
        audios = []

        for msg in msgs:
            for block in msg.get_content_blocks():
                if block["type"] == "text":
                    accumulated_text.append(f"{msg.name}: {block['text']}")

                elif block["type"] == "image":
                    source_type = block["source"]["type"]
                    if source_type == "url":
                        url = _to_openai_image_url(block["source"]["url"])

                    elif source_type == "base64":
                        data = block["source"]["data"]
                        media_type = block["source"]["media_type"]
                        url = f"data:{media_type};base64,{data}"

                    else:
                        raise ValueError(
                            f"Unsupported image source type: {source_type}",
                        )
                    images.append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": url,
                            },
                        },
                    )
                elif block["type"] == "audio":
                    input_audio = _to_openai_audio_data(block["source"])
                    audios.append(
                        {
                            "type": "input_audio",
                            "input_audio": input_audio,
                        },
                    )

        if accumulated_text:
            conversation_blocks.append(
                {"text": "\n".join(accumulated_text)},
            )

        if conversation_blocks:
            if conversation_blocks[0].get("text"):
                conversation_blocks[0]["text"] = (
                    conversation_history_prompt
                    + "<history>\n"
                    + conversation_blocks[0]["text"]
                )

            else:
                conversation_blocks.insert(
                    0,
                    {
                        "text": conversation_history_prompt + "<history>\n",
                    },
                )

            if conversation_blocks[-1].get("text"):
                conversation_blocks[-1]["text"] += "\n</history>"

            else:
                conversation_blocks.append({"text": "</history>"})

        conversation_blocks_text = "\n".join(
            conversation_block.get("text", "")
            for conversation_block in conversation_blocks
        )

        content_list: list[dict[str, Any]] = []
        if conversation_blocks_text:
            content_list.append(
                {
                    "type": "text",
                    "text": conversation_blocks_text,
                },
            )
        if images:
            content_list.extend(images)
        if audios:
            content_list.extend(audios)

        user_message = {
            "role": "user",
            "content": content_list,
        }

        if content_list:
            formatted_msgs.append(user_message)

        return formatted_msgs



# -*- coding: utf-8 -*-
"""The memory module."""

from ._in_memory_memory import InMemoryMemory
from ._long_term_memory_base import LongTermMemoryBase
from ._mem0_long_term_memory import Mem0LongTermMemory
from ._memory_base import MemoryBase
from ._reme import (
    ReMePersonalLongTermMemory,
    ReMeTaskLongTermMemory,
    ReMeToolLongTermMemory,
)

__all__ = [
    "MemoryBase",
    "InMemoryMemory",
    "LongTermMemoryBase",
    "Mem0LongTermMemory",
    "ReMePersonalLongTermMemory",
    "ReMeTaskLongTermMemory",
    "ReMeToolLongTermMemory",
]



# -*- coding: utf-8 -*-
"""The multi-modal-related tools module in agentscope."""
from ._dashscope_tools import (
    dashscope_image_to_text,
    dashscope_text_to_audio,
    dashscope_text_to_image,
)
from ._openai_tools import (
    openai_text_to_image,
    openai_edit_image,
    openai_text_to_audio,
    openai_create_image_variation,
    openai_image_to_text,
    openai_audio_to_text,
)

__all__ = [
    "dashscope_image_to_text",
    "dashscope_text_to_audio",
    "dashscope_text_to_image",
    "openai_text_to_image",
    "openai_text_to_audio",
    "openai_edit_image",
    "openai_create_image_variation",
    "openai_image_to_text",
    "openai_audio_to_text",
]



# -*- coding: utf-8 -*-
"""A file system based evaluator storage."""
import json
import os
from json import JSONDecodeError
from typing import Any, Callable

from ._evaluator_storage_base import EvaluatorStorageBase
from .._solution import SolutionOutput
from .._metric_base import MetricResult
from ...agent import AgentBase
from ...message import Msg


class FileEvaluatorStorage(EvaluatorStorageBase):
    """File system based evaluator storage, providing methods to save and
    retrieve evaluation results. So that the evaluation process can be resumed
    from the last saved state.

    The files are organized in a directory structure:
    - save_dir/
        - evaluation_result.json
        - evaluation_meta.json
        - {task_id}/
            - {repeat_id}/
                - solution.json
                - evaluation/
                    - {metric_name}.json
    """

    SOLUTION_FILE_NAME = "solution.json"
    EVALUATION_DIR_NAME = "evaluation"
    EVALUATION_RESULT_FILE = "evaluation_result.json"
    EVALUATION_META_FILE = "evaluation_meta.json"
    AGENT_PRINTING_LOG = "logging.txt"

    def __init__(self, save_dir: str) -> None:
        """Initialize the file evaluator storage."""
        self.save_dir = save_dir

    def _get_save_path(self, task_id: str, repeat_id: str, *args: str) -> str:
        """Get the save path for a given task and repeat ID."""
        return os.path.join(self.save_dir, repeat_id, task_id, *args)

    def save_solution_result(
        self,
        task_id: str,
        repeat_id: str,
        output: SolutionOutput,
        **kwargs: Any,
    ) -> None:
        """Save the solution result.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.
            output (`SolutionOutput`):
                The solution output to be saved.
        """
        path_file = self._get_save_path(
            task_id,
            repeat_id,
            self.SOLUTION_FILE_NAME,
        )
        os.makedirs(os.path.dirname(path_file), exist_ok=True)
        with open(path_file, "w", encoding="utf-8") as f:
            json.dump(output, f, ensure_ascii=False, indent=4)

    def save_evaluation_result(
        self,
        task_id: str,
        repeat_id: str,
        evaluation: MetricResult,
        **kwargs: Any,
    ) -> None:
        """Save the evaluation result.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.
            evaluation (`MetricResult`):
                The evaluation result to be saved.
        """
        path_file = self._get_save_path(
            task_id,
            repeat_id,
            self.EVALUATION_DIR_NAME,
            f"{evaluation.name}.json",
        )
        os.makedirs(os.path.dirname(path_file), exist_ok=True)
        with open(path_file, "w", encoding="utf-8") as f:
            json.dump(evaluation, f, ensure_ascii=False, indent=4)

    def get_evaluation_result(
        self,
        task_id: str,
        repeat_id: str,
        metric_name: str,
    ) -> MetricResult:
        """Get the evaluation result by the given task id and repeat id

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.
            metric_name (`str`):
                The metric name.

        Returns:
            `MetricResult`:
                The evaluation result for the given task and repeat ID.
        """
        path_file = self._get_save_path(
            task_id,
            repeat_id,
            self.EVALUATION_DIR_NAME,
            f"{metric_name}.json",
        )
        if not os.path.exists(path_file):
            raise FileNotFoundError(path_file)
        with open(path_file, "r", encoding="utf-8") as f:
            evaluation = json.load(f)
        return MetricResult(**evaluation)

    def get_solution_result(
        self,
        task_id: str,
        repeat_id: str,
        **kwargs: Any,
    ) -> SolutionOutput:
        """Get the solution result for the given task and repeat id from the
        file system.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.

        Raises:
            `FileNotFoundError`:
                If the solution result file does not exist for the given task
                and repeat ID.

        Returns:
            `SolutionOutput`:
                The solution output for the given task and repeat ID.
        """
        path_file = self._get_save_path(
            task_id,
            repeat_id,
            self.SOLUTION_FILE_NAME,
        )
        if not os.path.exists(path_file):
            raise FileNotFoundError(
                f"Solution result for task {task_id} and repeat {repeat_id} "
                "not found.",
            )

        try:
            with open(path_file, "r", encoding="utf-8") as f:
                solution_data = json.load(f)
        except JSONDecodeError as e:
            raise JSONDecodeError(
                f"Failed to load JSON from {path_file}: {e.msg}",
                e.doc,
                e.pos,
            ) from e

        return SolutionOutput(**solution_data)

    def solution_result_exists(self, task_id: str, repeat_id: str) -> bool:
        """Check if the solution for the given task and repeat is finished.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.

        Returns:
            `bool`:
                True if the solution result file exists, False otherwise.
        """
        path_file = self._get_save_path(
            task_id,
            repeat_id,
            self.SOLUTION_FILE_NAME,
        )

        return os.path.exists(path_file) and os.path.getsize(path_file) > 0

    def evaluation_result_exists(
        self,
        task_id: str,
        repeat_id: str,
        metric_name: str,
    ) -> bool:
        """Check if the evaluation result for the given solution and metric
        is finished.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.
            metric_name (`str`):
                The name of the metric.

        Returns:
            `bool`:
                True if the evaluation result file exists, False otherwise.
        """
        path_file = self._get_save_path(
            task_id,
            repeat_id,
            self.EVALUATION_DIR_NAME,
            f"{metric_name}.json",
        )
        return os.path.exists(path_file) and os.path.getsize(path_file) > 0

    def save_aggregation_result(
        self,
        aggregation_result: dict,
        **kwargs: Any,
    ) -> None:
        """Save the aggregation result.

        Args:
            aggregation_result (`dict`):
                A dictionary containing the aggregation result.
        """
        path_file = os.path.join(
            self.save_dir,
            self.EVALUATION_RESULT_FILE,
        )
        os.makedirs(os.path.dirname(path_file), exist_ok=True)
        with open(path_file, "w", encoding="utf-8") as f:
            json.dump(aggregation_result, f, ensure_ascii=False, indent=4)

    def aggregation_result_exists(
        self,
        **kwargs: Any,
    ) -> bool:
        """Check if the aggregation result exists

        Returns:
            `bool`:
                `True` if the aggregation result file exists.
        """
        path_file = os.path.join(
            self.save_dir,
            self.EVALUATION_RESULT_FILE,
        )
        return os.path.exists(path_file) and os.path.getsize(path_file) > 0

    def save_evaluation_meta(self, meta_info: dict) -> None:
        """Save the evaluation meta information.

        Args:
            meta_info (`dict`):
                A dictionary containing the meta information.
        """
        path_file = os.path.join(
            self.save_dir,
            self.EVALUATION_META_FILE,
        )
        os.makedirs(os.path.dirname(path_file), exist_ok=True)
        with open(path_file, "w", encoding="utf-8") as f:
            json.dump(meta_info, f, ensure_ascii=False, indent=4)

    def get_agent_pre_print_hook(
        self,
        task_id: str,
        repeat_id: str,
    ) -> Callable[[AgentBase, dict], None]:
        """Get a pre-print hook function for the agent to save the agent
        printing in the evaluation storage.

        Args:
            task_id (`str`):
                The task ID.
            repeat_id (`str`):
                The repeat ID for the task, usually the index of the repeat
                evaluation.

        Returns:
            `Callable[[AgentBase, dict], None]`:
                A hook function that takes an `AgentBase` instance and a
                keyword arguments dictionary as input, saving the agent's
                printing Msg into the evaluation storage.
        """

        def pre_print_hook(_agent: AgentBase, kwargs: dict) -> None:
            """Hook function to save agent's printing."""
            msg: Msg | None = kwargs.get("msg", None)
            last: bool = kwargs.get("last", False)

            if msg is None or not last:
                return

            # Only save the last message
            printing_str = []
            for block in msg.get_content_blocks():
                match block["type"]:
                    case "text":
                        printing_str.append(
                            f"{msg.name}: {block['text']}",
                        )
                    case "thinking":
                        printing_str.append(
                            f"{msg.name} (thinking): {block['text']}",
                        )
                    case _:
                        block_str = json.dumps(
                            block,
                            ensure_ascii=False,
                            indent=4,
                        )
                        if printing_str:
                            printing_str.append(block_str)
                        else:
                            printing_str.append(f"{msg.name}: {block_str}")

            path_file = self._get_save_path(
                task_id,
                repeat_id,
                self.AGENT_PRINTING_LOG,
            )
            os.makedirs(os.path.dirname(path_file), exist_ok=True)
            with open(path_file, "a", encoding="utf-8") as f:
                f.write("\n".join(printing_str) + "\n")

        return pre_print_hook



# -*- coding: utf-8 -*-
"""The dashscope API model classes."""
import collections
from datetime import datetime
from http import HTTPStatus
from typing import (
    Any,
    AsyncGenerator,
    Generator,
    Union,
    TYPE_CHECKING,
    List,
    Literal,
    Type,
)
from pydantic import BaseModel
from aioitertools import iter as giter

from ._model_base import ChatModelBase
from ._model_response import ChatResponse
from ._model_usage import ChatUsage
from .._utils._common import (
    _json_loads_with_repair,
    _create_tool_from_base_model,
)
from ..message import TextBlock, ToolUseBlock, ThinkingBlock
from ..tracing import trace_llm
from ..types import JSONSerializableObject
from .._logging import logger

if TYPE_CHECKING:
    from dashscope.api_entities.dashscope_response import GenerationResponse
    from dashscope.api_entities.dashscope_response import (
        MultiModalConversationResponse,
    )
else:
    GenerationResponse = (
        "dashscope.api_entities.dashscope_response.GenerationResponse"
    )
    MultiModalConversationResponse = (
        "dashscope.api_entities.dashscope_response."
        "MultiModalConversationResponse"
    )


class DashScopeChatModel(ChatModelBase):
    """The DashScope chat model class, which unifies the Generation and
    MultimodalConversation APIs into one method."""

    def __init__(
        self,
        model_name: str,
        api_key: str,
        stream: bool = True,
        enable_thinking: bool | None = None,
        generate_kwargs: dict[str, JSONSerializableObject] | None = None,
        base_http_api_url: str | None = None,
    ) -> None:
        """Initialize the DashScope chat model.

        Args:
            model_name (`str`):
                The model names.
            api_key (`str`):
                The dashscope API key.
            stream (`bool`):
                The streaming output or not
            enable_thinking (`bool | None`, optional):
                Enable thinking or not, only support Qwen3, QwQ, DeepSeek-R1.
                Refer to `DashScope documentation
                <https://help.aliyun.com/zh/model-studio/deep-thinking>`_
                for more details.
            generate_kwargs (`dict[str, JSONSerializableObject] | None`, \
            optional):
               The extra keyword arguments used in DashScope API generation,
               e.g. `temperature`, `seed`.
            base_http_api_url (`str | None`, optional):
                The base URL for DashScope API requests. If not provided,
                the default base URL from the DashScope SDK will be used.
        """
        if enable_thinking and not stream:
            logger.info(
                "In DashScope API, `stream` must be True when "
                "`enable_thinking` is True. ",
            )
            stream = True

        super().__init__(model_name, stream)

        self.api_key = api_key
        self.enable_thinking = enable_thinking
        self.generate_kwargs = generate_kwargs or {}

        if base_http_api_url is not None:
            import dashscope

            dashscope.base_http_api_url = base_http_api_url

    @trace_llm
    async def __call__(
        self,
        messages: list[dict[str, Any]],
        tools: list[dict] | None = None,
        tool_choice: Literal["auto", "none", "any", "required"]
        | str
        | None = None,
        structured_model: Type[BaseModel] | None = None,
        **kwargs: Any,
    ) -> ChatResponse | AsyncGenerator[ChatResponse, None]:
        """Get the response from the dashscope
        Generation/MultimodalConversation API by the given arguments.

        .. note:: We unify the dashscope generation and multimodal conversation
         APIs into one method, since they support similar arguments and share
         the same functionality.

        Args:
            messages (`list[dict[str, Any]]`):
                A list of dictionaries, where `role` and `content` fields are
                required.
            tools (`list[dict] | None`, default `None`):
                The tools JSON schemas that the model can use.
            tool_choice (`Literal["auto", "none", "any", "required"] | str \
             |  None`,  default `None`):
                Controls which (if any) tool is called by the model.
                 Can be "auto", "none", or specific tool name.
                 For more details, please refer to
                 https://help.aliyun.com/zh/model-studio/qwen-function-calling
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output. When provided, the model will be forced
                to return data that conforms to this schema by automatically
                converting the BaseModel to a tool function and setting
                `tool_choice` to enforce its usage. This enables structured
                output generation.

                .. note:: When `structured_model` is specified,
                    both `tools` and `tool_choice` parameters are ignored,
                    and the model will only perform structured output
                    generation without calling any other tools.

            **kwargs (`Any`):
                The keyword arguments for DashScope chat completions API,
                e.g. `temperature`, `max_tokens`, `top_p`, etc. Please
                refer to `DashScope documentation
                <https://help.aliyun.com/zh/dashscope/developer-reference/api-details>`_
                for more detailed arguments.
        """
        import dashscope

        # For qvq and qwen-vl models, the content field cannot be `None` or
        # `[{"text": None}]`, so we need to convert it to an empty list.
        if self.model_name.startswith("qvq") or "-vl" in self.model_name:
            for msg in messages:
                if msg["content"] is None or msg["content"] == [
                    {"text": None},
                ]:
                    msg["content"] = []

        kwargs = {
            "messages": messages,
            "model": self.model_name,
            "stream": self.stream,
            **self.generate_kwargs,
            **kwargs,
            "result_format": "message",
            # In agentscope, the `incremental_output` must be `True` when
            # `self.stream` is True
            "incremental_output": self.stream,
        }

        if tools:
            kwargs["tools"] = self._format_tools_json_schemas(tools)

        if tool_choice:
            self._validate_tool_choice(tool_choice, tools)
            kwargs["tool_choice"] = self._format_tool_choice(tool_choice)

        if (
            self.enable_thinking is not None
            and "enable_thinking" not in kwargs
        ):
            kwargs["enable_thinking"] = self.enable_thinking

        if structured_model:
            if tools or tool_choice:
                logger.warning(
                    "structured_model is provided. Both 'tools' and "
                    "'tool_choice' parameters will be overridden and "
                    "ignored. The model will only perform structured output "
                    "generation without calling any other tools.",
                )
            format_tool = _create_tool_from_base_model(structured_model)
            kwargs["tools"] = self._format_tools_json_schemas(
                [format_tool],
            )
            kwargs["tool_choice"] = self._format_tool_choice(
                format_tool["function"]["name"],
            )

        start_datetime = datetime.now()
        if self.model_name.startswith("qvq") or "-vl" in self.model_name:
            response = dashscope.MultiModalConversation.call(
                api_key=self.api_key,
                **kwargs,
            )

        else:
            response = await dashscope.aigc.generation.AioGeneration.call(
                api_key=self.api_key,
                **kwargs,
            )

        if self.stream:
            return self._parse_dashscope_stream_response(
                start_datetime,
                response,
                structured_model,
            )

        parsed_response = await self._parse_dashscope_generation_response(
            start_datetime,
            response,
            structured_model,
        )

        return parsed_response

    # pylint: disable=too-many-branches
    async def _parse_dashscope_stream_response(
        self,
        start_datetime: datetime,
        response: Union[
            AsyncGenerator[GenerationResponse, None],
            Generator[MultiModalConversationResponse, None, None],
        ],
        structured_model: Type[BaseModel] | None = None,
    ) -> AsyncGenerator[ChatResponse, Any]:
        """Given a DashScope streaming response generator, extract the content
            blocks and usages from it and yield ChatResponse objects.

        Args:
            start_datetime (`datetime`):
                The start datetime of the response generation.
            response (
                `Union[AsyncGenerator[GenerationResponse, None], Generator[ \
                MultiModalConversationResponse, None, None]]`
            ):
                DashScope streaming response generator (GenerationResponse or
                MultiModalConversationResponse) to parse.
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output.

        Returns:
            AsyncGenerator[ChatResponse, Any]:
                An async generator that yields ChatResponse objects containing
                the content blocks and usage information for each chunk in the
                streaming response.

        .. note::
            If `structured_model` is not `None`, the expected structured output
            will be stored in the metadata of the `ChatResponse`.
        """
        acc_content, acc_thinking_content = "", ""
        acc_tool_calls = collections.defaultdict(dict)
        metadata = None

        async for chunk in giter(response):
            if chunk.status_code != HTTPStatus.OK:
                raise RuntimeError(
                    f"Failed to get response from _ API: {chunk}",
                )

            message = chunk.output.choices[0].message

            # Update reasoning content
            if isinstance(message.get("reasoning_content"), str):
                acc_thinking_content += message["reasoning_content"]

            # Update text content
            if isinstance(message.content, str):
                acc_content += message.content
            elif isinstance(message.content, list):
                for item in message.content:
                    if isinstance(item, dict) and "text" in item:
                        acc_content += item["text"]

            # Update tool calls
            for tool_call in message.get("tool_calls", []):
                index = tool_call.get("index", 0)

                if "id" in tool_call and tool_call["id"] != acc_tool_calls[
                    index
                ].get("id"):
                    acc_tool_calls[index]["id"] = (
                        acc_tool_calls[index].get("id", "") + tool_call["id"]
                    )

                if "function" in tool_call:
                    func = tool_call["function"]
                    if "name" in func:
                        acc_tool_calls[index]["name"] = (
                            acc_tool_calls[index].get("name", "")
                            + func["name"]
                        )

                    if "arguments" in func:
                        acc_tool_calls[index]["arguments"] = (
                            acc_tool_calls[index].get("arguments", "")
                            + func["arguments"]
                        )

            # to content blocks
            content_blocks: list[TextBlock | ToolUseBlock | ThinkingBlock] = []
            if acc_thinking_content:
                content_blocks.append(
                    ThinkingBlock(
                        type="thinking",
                        thinking=acc_thinking_content,
                    ),
                )

            if acc_content:
                content_blocks.append(
                    TextBlock(
                        type="text",
                        text=acc_content,
                    ),
                )

            for tool_call in acc_tool_calls.values():
                repaired_input = _json_loads_with_repair(
                    tool_call.get("arguments", "{}") or "{}",
                )

                if not isinstance(repaired_input, dict):
                    repaired_input = {}

                content_blocks.append(
                    ToolUseBlock(
                        type="tool_use",
                        id=tool_call.get("id", ""),
                        name=tool_call.get("name", ""),
                        input=repaired_input,
                    ),
                )

                if structured_model:
                    metadata = repaired_input

            usage = None
            if chunk.usage:
                usage = ChatUsage(
                    input_tokens=chunk.usage.input_tokens,
                    output_tokens=chunk.usage.output_tokens,
                    time=(datetime.now() - start_datetime).total_seconds(),
                )

            parsed_chunk = ChatResponse(
                content=content_blocks,
                usage=usage,
                metadata=metadata,
            )
            yield parsed_chunk

    async def _parse_dashscope_generation_response(
        self,
        start_datetime: datetime,
        response: Union[
            GenerationResponse,
            MultiModalConversationResponse,
        ],
        structured_model: Type[BaseModel] | None = None,
    ) -> ChatResponse:
        """Given a DashScope GenerationResponse object, extract the content
        blocks and usages from it.

        Args:
            start_datetime (`datetime`):
                The start datetime of the response generation.
            response (
                `Union[GenerationResponse, MultiModalConversationResponse]`
            ):
                Dashscope GenerationResponse | MultiModalConversationResponse
                object to parse.
            structured_model (`Type[BaseModel] | None`, default `None`):
                A Pydantic BaseModel class that defines the expected structure
                for the model's output.

        Returns:
            ChatResponse (`ChatResponse`):
                A ChatResponse object containing the content blocks and usage.

        .. note::
            If `structured_model` is not `None`, the expected structured output
            will be stored in the metadata of the `ChatResponse`.
        """
        # Collect the content blocks from the response.
        if response.status_code != 200:
            raise RuntimeError(response)

        content_blocks: List[TextBlock | ToolUseBlock] = []
        metadata: dict | None = None

        message = response.output.choices[0].message
        content = message.get("content")

        if response.output.choices[0].message.get("content") not in [
            None,
            "",
            [],
        ]:
            if isinstance(content, list):
                for item in content:
                    if isinstance(item, dict) and "text" in item:
                        content_blocks.append(
                            TextBlock(
                                type="text",
                                text=item["text"],
                            ),
                        )
            else:
                content_blocks.append(
                    TextBlock(
                        type="text",
                        text=content,
                    ),
                )

        if message.get("tool_calls"):
            for tool_call in message["tool_calls"]:
                input_ = _json_loads_with_repair(
                    tool_call["function"].get(
                        "arguments",
                        "{}",
                    )
                    or "{}",
                )
                content_blocks.append(
                    ToolUseBlock(
                        type="tool_use",
                        name=tool_call["function"]["name"],
                        input=input_,
                        id=tool_call["id"],
                    ),
                )

                if structured_model:
                    metadata = input_

        # Usage information
        usage = None
        if response.usage:
            usage = ChatUsage(
                input_tokens=response.usage.input_tokens,
                output_tokens=response.usage.output_tokens,
                time=(datetime.now() - start_datetime).total_seconds(),
            )

        parsed_response = ChatResponse(
            content=content_blocks,
            usage=usage,
            metadata=metadata,
        )

        return parsed_response

    def _format_tools_json_schemas(
        self,
        schemas: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        """Format the tools JSON schema into required format for DashScope API.

        Args:
            schemas (`dict[str, dict[str, Any]]`):
                The tools JSON schemas.
        """
        # Check schemas format
        for value in schemas:
            if (
                not isinstance(value, dict)
                or "type" not in value
                or value["type"] != "function"
                or "function" not in value
            ):
                raise ValueError(
                    f"Each schema must be a dict with 'type' as 'function' "
                    f"and 'function' key, got {value}",
                )

        return schemas

    def _format_tool_choice(
        self,
        tool_choice: Literal["auto", "none", "any", "required"] | str | None,
    ) -> str | dict | None:
        """Format tool_choice parameter for API compatibility.

        Args:
            tool_choice (`Literal["auto", "none",  "any", "required"] | str \
            | None`, default  `None`):
                Controls which (if any) tool is called by the model.
                 Can be "auto", "none", or specific tool name.
                 For more details, please refer to
                 https://help.aliyun.com/zh/model-studio/qwen-function-calling
        Returns:
            `dict | None`:
                The formatted tool choice configuration dict, or None if
                    tool_choice is None.
        """
        if tool_choice is None:
            return None
        if tool_choice in ["auto", "none"]:
            return tool_choice
        if tool_choice in ["any", "required"]:
            logger.warning(
                "tool_choice '%s' is not supported by DashScope API. "
                "Supported options are 'auto', 'none', or specific function "
                "name. Automatically using 'auto' instead.",
                tool_choice,
            )
            return "auto"
        return {"type": "function", "function": {"name": tool_choice}}



# -*- coding: utf-8 -*-
"""The built-in hook functions in agentscope."""
from functools import partial

from ._studio_hooks import (
    as_studio_forward_message_pre_print_hook,
)
from .. import _config
from ..agent import AgentBase


__all__ = [
    "as_studio_forward_message_pre_print_hook",
]


def _equip_as_studio_hooks(
    studio_url: str,
) -> None:
    """Connect to the agentscope studio."""
    AgentBase.register_class_hook(
        "pre_print",
        "as_studio_forward_message_pre_print_hook",
        partial(
            as_studio_forward_message_pre_print_hook,
            studio_url=studio_url,
            run_id=_config.run_id,
        ),
    )



# -*- coding: utf-8 -*-
"""The knowledge base abstraction for retrieval-augmented generation (RAG)."""
from abc import abstractmethod
from typing import Any

from ._reader import Document
from ..embedding import EmbeddingModelBase
from ._store import VDBStoreBase
from ..message import TextBlock
from ..tool import ToolResponse


class KnowledgeBase:
    """The knowledge base abstraction for retrieval-augmented generation
    (RAG).

    The ``retrieve`` and ``add_documents`` methods need to be implemented
    in the subclasses. We also provide a quick method ``retrieve_knowledge``
    that enables the agent to retrieve knowledge easily.
    """

    embedding_store: VDBStoreBase
    """The embedding store for the knowledge base."""

    embedding_model: EmbeddingModelBase
    """The embedding model for the knowledge base."""

    def __init__(
        self,
        embedding_store: VDBStoreBase,
        embedding_model: EmbeddingModelBase,
    ) -> None:
        """Initialize the knowledge base."""
        self.embedding_store = embedding_store
        self.embedding_model = embedding_model

    @abstractmethod
    async def retrieve(
        self,
        query: str,
        limit: int = 5,
        score_threshold: float | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        """Retrieve relevant documents by the given query.

        Args:
            query (`str`):
                The query string to retrieve relevant documents.
            limit (`int`, defaults to 5):
                The number of relevant documents to retrieve.
            score_threshold (`float | None`, defaults to `None`):
                The score threshold to filter the retrieved documents. If
                provided, only documents with a score higher than the
                threshold will be returned.
            **kwargs (`Any`):
                Other keyword arguments for the vector database search API.
        """

    @abstractmethod
    async def add_documents(
        self,
        documents: list[Document],
        **kwargs: Any,
    ) -> None:
        """Add documents to the knowledge base, which will embed the documents
        and store them in the embedding store.

        Args:
            documents (`list[Document]`):
                A list of documents to add.
        """

    # A quick method that enable the agent to retrieve knowledge
    # Developers can wrap the `retrieve` method by themselves to support
    # more flexible usage
    async def retrieve_knowledge(
        self,
        query: str,
        limit: int = 5,
        score_threshold: float | None = None,
        **kwargs: Any,
    ) -> ToolResponse:
        """Retrieve relevant documents from the knowledge base. Note the
        `query` parameter is directly related to the retrieval quality, and
        for the same question, you can try many different queries to get the
        best results. Adjust the `limit` and `score_threshold` parameters
        to get more or fewer results.

        Args:
            query (`str`):
                The query string, which should be specific and concise. For
                example, you should provide the specific name instead of
                "you", "my", "he", "she", etc.
            limit (`int`, defaults to 3):
                The number of relevant documents to retrieve.
            score_threshold (`float`, defaults to 0.8):
                A threshold in [0, 1] and only the relevance score above this
                threshold will be returned. Reduce this value to get more
                results.
        """

        docs = await self.retrieve(
            query=query,
            limit=limit,
            score_threshold=score_threshold,
            **kwargs,
        )

        if len(docs):
            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"Score: {_.score}, "
                        f"Content: {_.metadata.content['text']}",
                    )
                    for _ in docs
                ],
            )
        return ToolResponse(
            content=[
                TextBlock(
                    type="text",
                    text="No relevant documents found. TRY to reduce the "
                    "`score_threshold` parameter to get "
                    "more results.",
                ),
            ],
        )



# -*- coding: utf-8 -*-
"""The base class for ReAct agent in agentscope."""
from abc import abstractmethod
from collections import OrderedDict
from typing import Callable, Any

from ._agent_base import AgentBase
from ._agent_meta import _ReActAgentMeta
from ..message import Msg


class ReActAgentBase(AgentBase, metaclass=_ReActAgentMeta):
    """The ReAct agent base class.

    To support ReAct algorithm, this class extends the AgentBase class by
    adding two abstract interfaces: reasoning and acting, while supporting
    hook functions at four positions: pre-reasoning, post-reasoning,
    pre-acting, and post-acting by the `_ReActAgentMeta` metaclass.
    """

    supported_hook_types: list[str] = [
        "pre_reply",
        "post_reply",
        "pre_print",
        "post_print",
        "pre_observe",
        "post_observe",
        "pre_reasoning",
        "post_reasoning",
        "pre_acting",
        "post_acting",
    ]
    """Supported hook types for the agent base class."""

    _class_pre_reasoning_hooks: dict[
        str,
        Callable[
            [
                "ReActAgentBase",  # self
                dict[str, Any],  # kwargs
            ],
            dict[str, Any] | None,  # The modified kwargs or None
        ],
    ] = OrderedDict()
    """The class-level pre-reasoning hooks, taking `self` object, the input
    arguments as input"""

    _class_post_reasoning_hooks: dict[
        str,
        Callable[
            [
                "ReActAgentBase",  # self
                dict[str, Any],  # kwargs
                Any,  # output
            ],
            Msg | None,  # the modified output message or None
        ],
    ] = OrderedDict()
    """The class-level post-reasoning hooks, taking `self` object, the input
    arguments and the output message as input, and return the modified output
    message or None if no modification is needed."""

    _class_pre_acting_hooks: dict[
        str,
        Callable[
            [
                "ReActAgentBase",  # self
                dict[str, Any],  # kwargs
            ],
            dict[str, Any] | None,  # The modified kwargs or None
        ],
    ] = OrderedDict()
    """The class-level pre-acting hooks, taking `self` object, the input
    arguments as input, and return the modified input arguments or None if no
    modification is needed."""

    _class_post_acting_hooks: dict[
        str,
        Callable[
            [
                "ReActAgentBase",  # self
                dict[str, Any],  # kwargs
                Any,  # output
            ],
            Msg | None,  # the modified output message or None
        ],
    ] = OrderedDict()
    """The class-level post-acting hooks, taking `self` object, the input
    arguments and the output message as input, and return the modified output
    message or None if no modification is needed."""

    def __init__(
        self,
    ) -> None:
        """Initialize the ReAct agent base class."""
        super().__init__()

        # Init reasoning and acting hooks
        self._instance_pre_reasoning_hooks = OrderedDict()
        self._instance_post_reasoning_hooks = OrderedDict()
        self._instance_pre_acting_hooks = OrderedDict()
        self._instance_post_acting_hooks = OrderedDict()

    @abstractmethod
    async def _reasoning(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """The reasoning process of the ReAct agent, which will be wrapped
        with pre- and post-hooks."""

    @abstractmethod
    async def _acting(self, *args: Any, **kwargs: Any) -> Any:
        """The acting process of the ReAct agent, which will be wrapped with
        pre- and post-hooks."""



# -*- coding: utf-8 -*-
"""The model usage class in agentscope."""
from dataclasses import dataclass, field
from typing import Literal

from .._utils._mixin import DictMixin


@dataclass
class ChatUsage(DictMixin):
    """The usage of a chat model API invocation."""

    input_tokens: int
    """The number of input tokens."""

    output_tokens: int
    """The number of output tokens."""

    time: float
    """The time used in seconds."""

    type: Literal["chat"] = field(default_factory=lambda: "chat")
    """The type of the usage, must be `chat`."""
